##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Lee-Ping Wang
# Contributors: Robert McGibbon and Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################

from __future__ import print_function, division
import os
import itertools
import numpy as np
from mdtraj.utils import cast_indices, in_units_of
from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils.six import string_types
from mdtraj.utils.six.moves import xrange

__all__ = ['ArcTrajectoryFile', 'load_arc']


##############################################################################
# Classes
##############################################################################

class _EOF(IOError):
    pass


@FormatRegistry.register_loader('.arc')
def load_arc(filename, stride=None, atom_indices=None, frame=None):
    """Load a TINKER .arc file from disk.

    Parameters
    ----------
    filename : path-like
        Path of TINKER .arc file.
    stride : int, default=None
        Only read every stride-th frame
    atom_indices : array_like, optional
        If not none, then read only a subset of the atoms coordinates from the
        file.
    frame : int, optional
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object.

    See Also
    --------
    mdtraj.ArcTrajectoryFile :  Low level interface to TINKER .arc files
    """
    from mdtraj.core.trajectory import _parse_topology

    if not isinstance(filename, (string_types, os.PathLike)):
        raise TypeError('filename must be of type path-like for load_arc. '
            'you supplied %s' % type(filename))

    atom_indices = cast_indices(atom_indices)

    with ArcTrajectoryFile(filename) as f:
        if frame is not None:
            f.seek(frame)
            n_frames = 1
        else:
            n_frames = None
        return f.read_as_traj(n_frames=n_frames, stride=stride,
                              atom_indices=atom_indices)


@FormatRegistry.register_fileobject('.arc')
class ArcTrajectoryFile(object):
    """Interface for reading and writing to an TINKER archive files.
    (Note that the TINKER .xyz format is identical to this.)  This is
    a file-like object, that both reading or writing depending on the
    `mode` flag. It implements the context manager protocol, so you
    can also use it with the python 'with' statement.

    The conventional units in the arc file is angstrom. The format only
    supports storing the cartesian coordinates and box lengths.

    Attributes
    ----------
    topology : Topology
        A single-chain, single-residue topology generated from the atom and bond
        information found in the TINKER archive/xyz file. It is only generated
        from the first member of the archive

    Parameters
    ----------
    filename : path-like
        The filename to open. A path to a file on disk.
    mode : {'r'}
        The mode in which to open the file, only 'r' for read is supported.
    force_overwrite : bool
        If opened in write mode, and a file by the name of `filename` already
        exists on disk, should we overwrite it?
    """

    distance_unit = 'angstroms'

    def __init__(self, filename, mode='r', force_overwrite=True):
        """Open an TINKER.arc file for reading/writing.
        """
        self._is_open = False
        self._filename = filename
        self._frame_index = 0
        self._mode = mode
        self.topology = None

        if mode == 'w':
            raise ValueError('Writing TINKER .arc files is not supported at this time')

        # track which line we're on. this is not essential, but its useful
        # when reporting errors to the user to say what line it occured on.
        self._line_counter = 0

        if mode == 'r':
            # if n_atoms is None:
            #     raise ValueError('To open a mdcrd file in mode="r", you must '
            #                      'supply the number of atoms, "n_atoms"')
            if not os.path.exists(filename):
                raise IOError("The file '%s' doesn't exist" % filename)
            self._fh = open(filename, 'r')
            self._is_open = True
        else:
            raise ValueError('mode must be "r". '
                             'you supplied "%s"' % mode)

    def seek(self, offset, whence=0):
        """Move to a new file position

        Parameters
        ----------
        offset : int
            A number of frames.
        whence : {0, 1, 2}
            0: offset from start of file, offset should be >=0.
            1: move relative to the current position, positive or negative
            2: move relative to the end of file, offset should be <= 0.
            Seeking beyond the end of a file is not supported
        """
        raise NotImplementedError()

    def tell(self):
        """Current file position

        Returns
        -------
        offset : int
            The current frame in the file.
        """
        raise NotImplementedError()

    def close(self):
        """Close the .arc file"""
        if self._is_open:
            self._fh.close()
            self._is_open = False

    def __del__(self):
        self.close()

    def __enter__(self):
        "Support the context manager protocol"
        return self

    def __exit__(self, *exc_info):
        "Support the context manager protocol"
        self.close()

    def __len__(self):
        "Number of frames in the file"
        raise NotImplementedError()

    def read_as_traj(self, n_frames=None, stride=None, atom_indices=None):
        """Read a trajectory from a ARC file

        Parameters
        ----------
        n_frames : int, optional
            If positive, then read only the next `n_frames` frames. Otherwise read all
            of the frames in the file.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        See Also
        --------
        read : Returns the raw data from the file
        """
        from mdtraj.core.trajectory import Trajectory
        if atom_indices is not None:
            topology = self.topology.subset(atom_indices)

        initial = int(self._frame_index)
        xyz, abc, ang = self.read(n_frames=n_frames, stride=stride, atom_indices=atom_indices)
        if len(xyz) == 0:
            return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), topology=topology)

        in_units_of(xyz, self.distance_unit, Trajectory._distance_unit, inplace=True)
        in_units_of(abc, self.distance_unit, Trajectory._distance_unit, inplace=True)

        if stride is None:
            stride = 1
        time = (stride*np.arange(len(xyz))) + initial

        return Trajectory(xyz=xyz, topology=self.topology, time=time,
                          unitcell_lengths=abc,
                          unitcell_angles=ang)

    def read(self, n_frames=None, stride=None, atom_indices=None):
        """Read data from a TINKER .arc file.

        Note that only the
        Cartesian coordinates are read in.  The .arc file also
        contains TINKER-specific numeric atom types and some bonding
        information, which we do not read in.

        Parameters
        ----------
        n_frames : int, None
            The number of frames you would like to read from the file.
            If None, all of the remaining frames will be loaded.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates
            from the file.

        Returns
        -------
        xyz : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=np.float32
            The cartesian coordinates, in angstroms
        """
        if not self._mode == 'r':
            raise ValueError('read() is only available when file is opened '
                             'in mode="r"')

        if n_frames is None:
            frame_counter = itertools.count()
        else:
            frame_counter = xrange(n_frames)

        if stride is None:
            stride = 1

        coords = []
        lengths = []
        angles = []
        for i in frame_counter:
            try:
                coord, length, angle = self._read()
                if atom_indices is not None:
                    coord = coord[atom_indices, :]
            except _EOF:
                break

            coords.append(coord)
            lengths.append(length)
            angles.append(angle)

            for j in range(stride - 1):
                # throw away these frames
                self._read()

        coords = np.array(coords)
        if any(l is None for l in lengths):
            lengths = angles = None
        else:
            lengths = np.array(lengths)
            angles = np.array(angles)

        return coords, lengths, angles

    def _read(self):
        "Read a single frame"
        from mdtraj.core.topology import Topology
        from mdtraj.core.element import Element, virtual
        # Read in the number of atoms.
        line = self._fh.readline()
        if line == '':
            raise _EOF()

        self._n_atoms = int(line.split()[0])
        self._line_counter += 1

        coords = np.empty((self._n_atoms, 3), dtype=np.float32)
        bond_partners = [[] for i in xrange(self._n_atoms)]
        atom_names = ['' for i in xrange(self._n_atoms)]
        line = self._fh.readline()
        s = line.split()
        self._line_counter += 1
        # See if we have box info on this line or not
        cell_lengths = cell_angles = None
        if len(s) == 6:
            try:
                cell_lengths = np.asarray(
                                [float(s[0]), float(s[1]), float(s[2])]
                )
                cell_angles = np.asarray(
                                [float(s[3]), float(s[4]), float(s[5])]
                )
                line = self._fh.readline()
                s = line.split()
                self._line_counter += 1
            except ValueError:
                pass
        i = 0
        while i < self._n_atoms - 1:
            atom_names[i] = s[1]
            bond_partners[i] = [int(x) for x in s[6:]]
            coords[i,:] = [float(s[pos]) for pos in [2, 3, 4]]
            i += 1
            line = self._fh.readline()
            s = line.split()
            self._line_counter += 1
        # Now do the last atom
        atom_names[i] = s[1]
        bond_partners[i] = [int(x) for x in s[6:]]
        coords[i,:] = [float(s[pos]) for pos in [2, 3, 4]]
        # Now see if we have to build a topology
        if self.topology is None:
            self.topology = top = Topology()
            chain = top.add_chain()                # only 1 chain
            res = top.add_residue('RES', chain, 1) # only 1 residue
            for at in atom_names:
                # First get the element. Try for common 2-letter elements, then
                # use the first letter only (default to None if I can't find it)
                if at[:2].upper() in ('NA', 'CL', 'MG'):
                    elem = Element.getBySymbol(at[:2])
                else:
                    try:
                        elem = Element.getBySymbol(at[0])
                    except KeyError:
                        elem = virtual
                top.add_atom(at, elem, res)
            # Now add the bonds
            atoms = list(top.atoms)
            for i, bonds in enumerate(bond_partners):
                me = atoms[i]
                for b in bonds:
                    b -= 1
                    if b < i: continue
                    it = atoms[b]
                    top.add_bond(me, it)

        self._frame_index += 1
        return coords, cell_lengths, cell_angles

    def write(self, xyz):
        """ The ArcTrajectoryFile does not have a write method,
        because TINKER .arc files have special numerical atom types
        which are not shared by any other trajectory file format.

        Parameters
        ----------
        xyz : np.ndarray, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of the atoms to write.

        """

        raise RuntimeError('write() is not available for .arc files')




cdef extern from "include/xdrfile.h":
    ctypedef struct XDRFILE:
        pass

    XDRFILE* xdrfile_open (char * path, char * mode)
    int xdrfile_close (XDRFILE * xfp)
    ctypedef float matrix[3][3]
    ctypedef float rvec[3]


cdef extern from "include/xdrfile_trr.h":
    int read_trr_natoms(char *fn, int *natoms)
    int read_trr_nframes(char* fn, unsigned long *nframes)

    # Read one frame of an open xtc file. If either of x, v, f, box are
    # NULL the arrays will be read from the file but not used.
    int read_trr(XDRFILE *xd, int natoms, int *step, float *t, float* lambd,
        matrix box, rvec* x, rvec* v, rvec* f)

    # Write a frame to xtc file
    int write_trr(XDRFILE *xd, int natoms, int step, float t, float lambd,
        matrix box, rvec* x, rvec* v, rvec* f)

cdef extern from "include/trr_header.h":
    ctypedef struct t_trnheader:
        int v_size
        int f_size

    # header
    int do_trnheader(XDRFILE *xd, int bRead, t_trnheader *sh)



from __future__ import print_function, division
from .hdf5reporter import HDF5Reporter
from .netcdfreporter import NetCDFReporter
from .dcdreporter import DCDReporter
from .xtcreporter import XTCReporter



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
# 
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

"""Sanity checking for molecular dynamics trajectories. This script is currently
a work in progress. Contributions are encouraged. 
"""
#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from __future__ import print_function
import os
import sys
import warnings
import functools
import operator
from argparse import ArgumentParser

import numpy as np
import mdtraj as md
from mdtraj.utils import import_, ilen
from mdtraj.geometry.internal import COVALENT_RADII
from mdtraj.core.trajectory import _parse_topology

spatial = import_('scipy.spatial')

#------------------------------------------------------------------------------
# Code
#------------------------------------------------------------------------------

class NoTopologyError(Exception):
    def __init__(self):
        super(NoTopologyError, self).__init__("One more more of the "
            "trajectory files should contain topology information (i.e. "
            "either HDF5 or PDB)")


def parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('files', nargs='+', help='''Input trajectory file(s),
        in any supported format.''')
    # parser.add_argument('-n', '--noload', action='store_true', help='''Do not load the coordinate data from the trajectory, for example if the trajectory is too large to load into memory. Only a limited number of checks will be done.''')
    parser.add_argument('-t', '--topology', type=str, help='''Topology for the system (.prmtop/.pdb)''')
    parser.add_argument('--bond-low', type=float, help='''Minimum fraction of sum of covalent radii for bonded atoms. Default=0.4''', default=0.4)
    parser.add_argument('--bond-high', type=float, help='''Maximum fraction of sum of covalent radii for bonded atoms. Default=1.2''', default=1.2)
    parser.add_argument('--rmsd-tolerance', type=float, help='''Maximum tolerance for percent change in RMSD. Default=100.0''', default=100.0)
    return parser.parse_args(), parser


def main(args, parser):
    inspector = Inspector(args.bond_low, args.bond_high, args.rmsd_tolerance)

    topology_files = [fn for fn in args.files if os.path.splitext(fn)[1] in ['.pdb', '.prmtop', '.h5']]
    if args.topology is None and len(topology_files) > 0:
        args.topology = topology_files[0]

    inspector.load_topology(args.topology)

    for f in args.files:
        if not os.path.exists(f):
            parser.error("File '%s' does not exist" % f)
        if not os.path.isfile(f):
            parser.error("File '%s' is not a file" % f)

        inspector.load_trajectory(f)


class Inspector(object):
    def __init__(self, bond_low, bond_high, rmsd_tolerance):
        self._printed_section = False
        self.t = None
        self.fn = None

        self.bond_low = bond_low
        self.bond_high = bond_high
        self.rmsd_tolerance = rmsd_tolerance

    def load_topology(self, fn):
        self.fn = fn
        self.topology = _parse_topology(fn)
        self.check_topology()

    def load_trajectory(self, fn):
        self.fn = fn

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.t = md.load(fn, top=self.topology)

        self.check_shape()
        self.check_unitcell()
        self.check_bonds()
        self.check_nonbonded()
        self.check_imaging()

    def check_shape(self):
        self.section('Shape')
        self.log('Number of frames: %d' % self.t.xyz.shape[0])
        self.log('Number of atoms:  %d' % self.t.xyz.shape[1])

    def check_unitcell(self):
        self.section('Unitcell')
        if not self.t._have_unitcell:
            self.log('No unitcell information')
        else:
            self.log('First frame:')
            self.log('Unitcell angles (deg): %s' % self.t.unitcell_angles[0])
            self.log('Unitcell lengths (nm): %s' % self.t.unitcell_lengths[0])

    def check_topology(self):
        self.section('topology')
        self.log('Number of Atoms:    %d' % ilen(self.topology.atoms))
        self.log('Number of Residues: %d' % ilen(self.topology.residues))
        self.log('Number of Chains:   %d' % ilen(self.topology.chains))
        self.log('Residues:           %s' % ', '.join(['%s (%d atoms)' % (r, ilen(r.atoms)) for r in self.topology.residues]))
        self.log('Unique atom names:  %s' % ', '.join(np.unique([a.name for a in self.topology.atoms])))

    def check_bonds(self):
        self.section("Bond Check (without PBCs)")

        radii = []
        pairs = []
        for (a, b) in self.topology.bonds:
            try:
                radsum = COVALENT_RADII[a.element.symbol] + COVALENT_RADII[b.element.symbol]
            except KeyError:
                raise NotImplementedError("I don't have radii information for all of your atoms")
            radii.append(radsum)
            pairs.append((a.index, b.index))

        radii = np.array(radii)
        pairs = np.array(pairs)

        distances = md.compute_distances(self.t, pairs, periodic=False)
        low, high = self.bond_low * radii, self.bond_high * radii
        extreme = np.logical_or(distances < low, distances > high)

        if np.any(extreme):
            frames, bonds = np.nonzero(extreme)
            frame, bond = frames[0], bonds[0]
            a1 = self.topology.atom(pairs[bond][0])
            a2 = self.topology.atom(pairs[bond][0])

            self.log('error: atoms (%s) and (%s) are bonded according to the topology ' % (a1, a2))
            self.log('but they are a distance of %.3f nm apart in frame %d' % (distances[frame, bond], frame))
        else:
            self.log("All good.")

    def section(self, title):
        if self._printed_section:
            print()
        print('=== %s: %s ===' % (os.path.basename(self.fn), title))
        self._printed_section = True

    def log(self, msg):
        print(msg)

    def check_nonbonded(self):
        self.section('Nonbonded Check (without PBCs)')

        # which atoms have a nonbonded interaction. exluce atoms interacting
        # with themselves or with atoms they're bonded to.
        nonbond_mask = np.logical_not(np.eye(self.t.n_atoms, dtype=np.bool))
        for (a, b) in self.t.topology.bonds:
            nonbond_mask[a.index, b.index] = False
            nonbond_mask[b.index, a.index] = False

        for i in range(min(5, self.t.n_frames)):
            dist = spatial.distance_matrix(self.t.xyz[i], self.t.xyz[i])

            dist = np.ma.array(dist, mask=np.logical_not(nonbond_mask))
            a, b = np.unravel_index(np.ma.argmin(dist), dist.shape)
            names = [q.residue.name + ' ' + q.name for q in self.t.topology.atoms]
            self.log('Frame %d: closest nb dist between '
                     '%d (%s), %d (%s), at d=%.4f nm' % (i, a, names[a], b, names[b], dist[a, b]))

    def check_imaging(self):
        if self.t.n_frames > 2:
            self.section('Imaging')

            r = md.rmsd(target=self.t, reference=self.t[0])
            percent_change = np.divide(np.abs(r[2:]-r[1:-1]), r[1:-1]) * 100.0
            if np.any(percent_change > self.rmsd_tolerance):
                self.log('Potential imaging issue: %s' % self.fn)
            else:
                self.log('No imaging issue detected')


def entry_point():
    args, parser = parse_args()
    main(args, parser)


if __name__ == '__main__':
    entry_point()



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2016 Stanford University and the Authors
#
# Authors: Peter Eastman
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################

from __future__ import print_function, division
import numpy as np
from mdtraj.utils import ensure_type

from libcpp.vector cimport vector

__all__ = ['compute_neighborlist']

cdef extern from "neighborlist.h":
    vector[vector[int]] _compute_neighborlist(float* positions, int n_atoms, float cutoff, float* box_matrix) nogil


##############################################################################
# Functions
##############################################################################

def compute_neighborlist(traj, cutoff, frame=0, periodic=True):
    """compute_neighborlist(traj, cutoff, frame=0, periodic=True)

    Find (spatially) neighboring atoms in a trajectory.

    For every atom, build a list of all other atoms that are within a cutoff
    distance of it.

    Parameters
    ----------
    traj : md.Trajectory
        An MDTraj trajectory
    cutoff : float
        Distance cutoff to define 'neighboring'
    frame : int
        Index of the frame of the trajectory to build the neighbor list based on
    periodic : bool
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.

    Returns
    -------
    neighbors : list of np.ndarray, dtype=int
        List of arrays, of length n_atoms. neighbors[i] contains the indices of all
        atoms within the cutoff distance of atom i.
    """

    cdef float[:, ::1] xyz = traj.xyz[frame]
    cdef float[:, ::1] box_matrix
    cdef float* box_matrix_pointer
    cdef int is_periodic = periodic and (traj.unitcell_vectors is not None)
    if is_periodic:
        unitcell_vectors = ensure_type(traj.unitcell_vectors[frame],
                                       dtype=np.float32, ndim=2,
                                       name='unitcell_vectors',
                                       warn_on_cast=False)
        box_matrix = np.asarray(unitcell_vectors, order='c')
        box_matrix_pointer = &box_matrix[0,0]
    else:
        box_matrix_pointer = NULL
    cdef int n_atoms = traj.xyz.shape[1]
    cdef vector[vector[int]] atom_neighbors = _compute_neighborlist(&xyz[0,0], n_atoms, cutoff, box_matrix_pointer)

    # Copy the results over to Numpy arrays.

    cdef int[::1] atom_neighbors_mview
    cdef list neighbors = []
    cdef int i
    for i in range(n_atoms):
        if atom_neighbors[i].size() > 0:
            atom_neighbors_mview = <int[:atom_neighbors[i].size()]> (<int*> (&atom_neighbors[i][0]))
            neighbors.append(np.array(atom_neighbors_mview, dtype=int, copy=True))
        else:
            neighbors.append(np.empty(0, dtype=int))
    return neighbors



"""
Execute each notebook as a test, reporting an error if any cell throws an exception.
Adapted from https://gist.github.com/minrk/2620876.
"""
from __future__ import print_function
import os
import sys
import socket
from distutils.spawn import find_executable as _find_executable

import nbformat
import pytest
from jupyter_client import KernelManager
from six.moves.queue import Empty

FLAKEY_LIST = ['centroids.ipynb', 'native-contact.ipynb', 'hbonds.ipynb']
SPARTA_PLUS = ['sparta+', 'SPARTA+', 'SPARTA+.linux']
TIMEOUT = 60  # seconds

test_dir = os.path.dirname(os.path.abspath(__file__))
examples = [pytest.param(fn, marks=pytest.mark.flaky) if fn in FLAKEY_LIST else fn
            for fn in os.listdir(test_dir) if fn.endswith('.ipynb')]


def is_network_connected():
    try:
        # connect to the host -- tells us if the host is actually
        # reachable
        socket.create_connection(("1.1.1.1", 53))
        return True
    except OSError:
        pass
    return False


def find_executable(names):
    for possible in names:
        result = _find_executable(possible)
        if result is not None:
            return result
    return None


@pytest.fixture(params=examples)
def example_fn(request):
    if 'openmm' in request.param:
        try:
            from simtk.openmm import app
        except ImportError:
            pytest.skip("Openmm required for example notebook `{}`".format(request.param))

    if "nmr" in request.param:
        if find_executable(SPARTA_PLUS) is None:
            pytest.skip("Sparta+ not found for example notebook `{}`".format(request.param))

    if not is_network_connected():
        if any(x in request.param for x in ("native-contact", "hbonds")):
            pytest.skip("Network access required")

    cwd = os.path.abspath('.')
    os.chdir(test_dir)
    yield request.param
    os.chdir(cwd)


def test_example_notebook(example_fn):
    with open(example_fn) as f:
        nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)
    run_notebook(nb)


def run_notebook(nb):
    km = KernelManager()
    km.start_kernel(stderr=open(os.devnull, 'w'))
    kc = km.client()
    kc.start_channels()
    # simple ping:
    kc.execute("pass")
    kc.get_shell_msg()

    failures = 0
    for cell in nb.cells:
        if cell.cell_type != 'code':
            continue
        kc.execute(cell.source)
        try:
            # wait for finish, w/ timeout
            reply = kc.get_shell_msg(timeout=TIMEOUT)['content']
        except Empty:
            raise Exception(
                'Timeout (%.1f) when executing the following %s cell: "%s"' %
                (TIMEOUT, cell.cell_type, cell.source.strip()))
        if reply['status'] == 'error':
            failures += 1
            print("\nFAILURE:", file=sys.stderr)
            print('\n'.join(reply['traceback']), file=sys.stderr)
            print(file=sys.stderr)

    kc.stop_channels()
    km.shutdown_kernel()
    del km
    if failures > 0:
        raise Exception()



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Christopher M. Bruns
# Contributors: Robert McGibbon, Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


"""
element.py: Used for managing elements.

This is part of the OpenMM molecular simulation toolkit originating from
Simbios, the NIH National Center for Physics-Based Simulation of
Biological Structures at Stanford, funded under the NIH Roadmap for
Medical Research, grant U54 GM072970. See https://simtk.org.

Portions copyright (c) 2012 Stanford University and the Authors.
Authors: Christopher M. Bruns
Contributors: Robert T. McGibbon

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import print_function, division
import numpy as np
from mdtraj.utils.unit.quantity import is_quantity
from mdtraj.utils.unit.unit_definitions import daltons

class Element(tuple):
    """An Element represents a chemical element.

    The mdtraj.pdb.element module contains objects for all the standard chemical elements,
    such as element.hydrogen or element.carbon.  You can also call the static method
    Element.getBySymbol() to look up the Element with a particular chemical symbol."""
    __slots__ = []
    _elements_by_symbol = {}
    _elements_by_atomic_number = {}

    def __new__(cls, number, name, symbol, mass, radius):
        """Create a new element

        Parameters
        ----------
        number : int
            The atomic number of the element
        name : str
            The name of the element
        symbol : str
            The chemical symbol of the element
        mass : float
            The atomic mass of the element
        radius : float
            The van der Waals radius of the element, in nm.
        """

        newobj = tuple.__new__(cls, (number, name, symbol, mass, radius))

        # Index this element in a global table
        s = symbol.strip().upper()
        assert s not in Element._elements_by_symbol
        Element._elements_by_symbol[s] = newobj
        if number in Element._elements_by_atomic_number:
            other_element = Element._elements_by_atomic_number[number]
            if mass < other_element.mass:
                # If two "elements" share the same atomic number, they're
                # probably hydrogen and deuterium, and we want to choose
                # the lighter one to put in the table by atomic_number,
                # since it's the "canonical" element.
                Element._elements_by_atomic_number[number] = newobj
        else:
            Element._elements_by_atomic_number[number] = newobj

        return newobj

    def __reduce__(self):
        # __reduce__ is part of the pickle protocol. we need to make sure that
        # elements still act as singletons after a pickle load/save cycle --
        # so load() has to *not create* a new object.
        # see http://docs.python.org/3.3/library/pickle.html#object.__reduce__

        # relevant test:
        # >>> cPickle.loads(cPickle.dumps(md.load(get_fn('bpti.pdb')).topology))
        return str(self.name)

    @staticmethod
    def getBySymbol(symbol):
        """Get the Element with a particular chemical symbol

        Parameters
        ----------
        symbol : str

        Returns
        -------
        element : Element
        """
        s = symbol.strip().upper()
        return Element._elements_by_symbol[s]

    @staticmethod
    def getByAtomicNumber(number):
        """ Get the element with a particular atomic number

        Parameters
        ----------
        number : int

        Returns
        -------
        element : Element
        """
        return Element._elements_by_atomic_number[number]

    @staticmethod
    def getByMass(mass):
        """Get the element whose mass is CLOSEST to the requested mass. This
        method should not be used for repartitioned masses.

        Parameters
        ----------
        mass : float

        Returns
        -------
        element : Element
        """
        # Convert any masses to daltons
        if is_quantity(mass):
            mass = mass.value_in_unit(daltons)
        diff = mass
        best_guess = None

        for key in Element._elements_by_atomic_number:
            element = Element._elements_by_atomic_number[key]
            massdiff = abs(element.mass - mass)
            if massdiff < diff:
                best_guess = element
                diff = massdiff

        return best_guess

    @property
    def number(self):
        """Atomic number."""
        return tuple.__getitem__(self, 0)

    @property
    def name(self):
        """Element name"""
        return tuple.__getitem__(self, 1)

    @property
    def symbol(self):
        """Element symbol"""
        return tuple.__getitem__(self, 2)

    @property
    def mass(self):
        """Element mass"""
        return tuple.__getitem__(self, 3)

    @property
    def radius(self):
        """Element atomic radius

        van der Waals radii are taken from A. Bondi, J. Phys. Chem., 68, 441 -
        452, 1964, except the value for H, which is taken from R.S.
        Rowland & R. Taylor, J.Phys.Chem., 100, 7384 - 7391, 1996. Radii
        that are not available in either of these publications have RvdW =
        2.00 A. The radii for Ions (Na, K, Cl, Ca, Mg, and Cs are based on
        the CHARMM27 Rmin/2 parameters for (SOD, POT, CLA, CAL, MG, CES) by
        default.
        """
        return tuple.__getitem__(self, 4)

    def __getitem__(self, item):
        raise TypeError

    def __str__(self):
        return self.name

    @property
    def atomic_number(self):
        """Atomic number"""
        return tuple.__getitem__(self, 0)

    # Make it so only virtual sites evaluate to boolean False (since it's really
    # *not* an element)
    def __bool__(self):
        return bool(self.mass)

    def __nonzero__(self):
        return bool(self.mass)



# This is for backward compatibility.
def get_by_symbol(symbol):
    s = symbol.strip().upper()
    return Element._elements_by_symbol[s]



# van der Waals radii are taken from A. Bondi,
# J. Phys. Chem., 68, 441 - 452, 1964,
# except the value for H, which is taken from R.S. Rowland & R. Taylor,
# J.Phys.Chem., 100, 7384 - 7391, 1996. Radii that are not available in
# either of these publications have RvdW = 2.00 A
# The radii for Ions (Na, K, Cl, Ca, Mg, and Cs are based on the CHARMM27
# Rmin/2 parameters for (SOD, POT, CLA, CAL, MG, CES) by default.

virtual =        Element(  0,"virtual_site","VS", 0.0, 0.0)
hydrogen =       Element(  1,"hydrogen","H", 1.007947, 0.12)
deuterium =      Element(  1,"deuterium","D", 2.0135532127, 0.12)
helium =         Element(  2,"helium","He", 4.003, 0.14)
lithium =        Element(  3,"lithium","Li", 6.9412, 0.182)
beryllium =      Element(  4,"beryllium","Be", 9.0121823, 0.2)
boron =          Element(  5,"boron","B", 10.8117, 0.2)
carbon =         Element(  6,"carbon","C", 12.01078, 0.17)
nitrogen =       Element(  7,"nitrogen","N", 14.00672, 0.155)
oxygen =         Element(  8,"oxygen","O", 15.99943, 0.152)
fluorine =       Element(  9,"fluorine","F", 18.99840325, 0.147)
neon =           Element( 10,"neon","Ne", 20.17976, 0.154)
sodium =         Element( 11,"sodium","Na", 22.989769282, 0.136)
magnesium =      Element( 12,"magnesium","Mg", 24.30506, 0.118)
aluminum =       Element( 13,"aluminum","Al", 26.98153868, 0.2)
silicon =        Element( 14,"silicon","Si", 28.08553, 0.21)
phosphorus =     Element( 15,"phosphorus","P", 30.9737622, 0.18)
sulfur =         Element( 16,"sulfur","S", 32.0655, 0.18)
chlorine =       Element( 17,"chlorine","Cl", 35.4532, 0.227)
argon =          Element( 18,"argon","Ar", 39.9481, 0.188)
potassium =      Element( 19,"potassium","K", 39.09831, 0.176)
calcium =        Element( 20,"calcium","Ca", 40.0784, 0.137)
scandium =       Element( 21,"scandium","Sc", 44.9559126, 0.2)
titanium =       Element( 22,"titanium","Ti", 47.8671, 0.2)
vanadium =       Element( 23,"vanadium","V", 50.94151, 0.2)
chromium =       Element( 24,"chromium","Cr", 51.99616, 0.2)
manganese =      Element( 25,"manganese","Mn", 54.9380455, 0.2)
iron =           Element( 26,"iron","Fe", 55.8452, 0.2)
cobalt =         Element( 27,"cobalt","Co", 58.9331955, 0.2)
nickel =         Element( 28,"nickel","Ni", 58.69342, 0.163)
copper =         Element( 29,"copper","Cu", 63.5463, 0.14)
zinc =           Element( 30,"zinc","Zn", 65.4094, 0.139)
gallium =        Element( 31,"gallium","Ga", 69.7231, 0.107)
germanium =      Element( 32,"germanium","Ge", 72.641, 0.2)
arsenic =        Element( 33,"arsenic","As", 74.921602, 0.185)
selenium =       Element( 34,"selenium","Se", 78.963, 0.19)
bromine =        Element( 35,"bromine","Br", 79.9041, 0.185)
krypton =        Element( 36,"krypton","Kr", 83.7982, 0.202)
rubidium =       Element( 37,"rubidium","Rb", 85.46783, 0.2)
strontium =      Element( 38,"strontium","Sr", 87.621, 0.2)
yttrium =        Element( 39,"yttrium","Y", 88.905852, 0.2)
zirconium =      Element( 40,"zirconium","Zr", 91.2242, 0.2)
niobium =        Element( 41,"niobium","Nb", 92.906382, 0.2)
molybdenum =     Element( 42,"molybdenum","Mo", 95.942, 0.2)
technetium =     Element( 43,"technetium","Tc", 98, 0.2)
ruthenium =      Element( 44,"ruthenium","Ru", 101.072, 0.2)
rhodium =        Element( 45,"rhodium","Rh", 102.905502, 0.2)
palladium =      Element( 46,"palladium","Pd", 106.421, 0.163)
silver =         Element( 47,"silver","Ag", 107.86822, 0.172)
cadmium =        Element( 48,"cadmium","Cd", 112.4118, 0.158)
indium =         Element( 49,"indium","In", 114.8183, 0.193)
tin =            Element( 50,"tin","Sn", 118.7107, 0.217)
antimony =       Element( 51,"antimony","Sb", 121.7601, 0.2)
tellurium =      Element( 52,"tellurium","Te", 127.603, 0.206)
iodine =         Element( 53,"iodine","I", 126.904473, 0.198)
xenon =          Element( 54,"xenon","Xe", 131.2936, 0.216)
cesium =         Element( 55,"cesium","Cs", 132.90545192, 0.21)
barium =         Element( 56,"barium","Ba", 137.3277, 0.2)
lanthanum =      Element( 57,"lanthanum","La", 138.905477, 0.2)
cerium =         Element( 58,"cerium","Ce", 140.1161, 0.2)
praseodymium =   Element( 59,"praseodymium","Pr", 140.907652, 0.2)
neodymium =      Element( 60,"neodymium","Nd", 144.2423, 0.2)
promethium =     Element( 61,"promethium","Pm", 145, 0.2)
samarium =       Element( 62,"samarium","Sm", 150.362, 0.2)
europium =       Element( 63,"europium","Eu", 151.9641, 0.2)
gadolinium =     Element( 64,"gadolinium","Gd", 157.253, 0.2)
terbium =        Element( 65,"terbium","Tb", 158.925352, 0.2)
dysprosium =     Element( 66,"dysprosium","Dy", 162.5001, 0.2)
holmium =        Element( 67,"holmium","Ho", 164.930322, 0.2)
erbium =         Element( 68,"erbium","Er", 167.2593, 0.2)
thulium =        Element( 69,"thulium","Tm", 168.934212, 0.2)
ytterbium =      Element( 70,"ytterbium","Yb", 173.043, 0.2)
lutetium =       Element( 71,"lutetium","Lu", 174.9671, 0.2)
hafnium =        Element( 72,"hafnium","Hf", 178.492, 0.2)
tantalum =       Element( 73,"tantalum","Ta", 180.947882, 0.2)
tungsten =       Element( 74,"tungsten","W", 183.841, 0.2)
rhenium =        Element( 75,"rhenium","Re", 186.2071, 0.2)
osmium =         Element( 76,"osmium","Os", 190.233, 0.2)
iridium =        Element( 77,"iridium","Ir", 192.2173, 0.2)
platinum =       Element( 78,"platinum","Pt", 195.0849, 0.172)
gold =           Element( 79,"gold","Au", 196.9665694, 0.166)
mercury =        Element( 80,"mercury","Hg", 200.592, 0.155)
thallium =       Element( 81,"thallium","Tl", 204.38332, 0.196)
lead =           Element( 82,"lead","Pb", 207.21, 0.202)
bismuth =        Element( 83,"bismuth","Bi", 208.980401, 0.2)
polonium =       Element( 84,"polonium","Po", 209, 0.2)
astatine =       Element( 85,"astatine","At", 210, 0.2)
radon =          Element( 86,"radon","Rn", 222.018, 0.2)
francium =       Element( 87,"francium","Fr", 223, 0.2)
radium =         Element( 88,"radium","Ra", 226, 0.2)
actinium =       Element( 89,"actinium","Ac", 227, 0.2)
thorium =        Element( 90,"thorium","Th", 232.038062, 0.2)
protactinium =   Element( 91,"protactinium","Pa", 231.035882, 0.2)
uranium =        Element( 92,"uranium","U", 238.028913, 0.186)
neptunium =      Element( 93,"neptunium","Np", 237, 0.2)
plutonium =      Element( 94,"plutonium","Pu", 244, 0.2)
americium =      Element( 95,"americium","Am", 243, 0.2)
curium =         Element( 96,"curium","Cm", 247, 0.2)
berkelium =      Element( 97,"berkelium","Bk", 247, 0.2)
californium =    Element( 98,"californium","Cf", 251, 0.2)
einsteinium =    Element( 99,"einsteinium","Es", 252, 0.2)
fermium =        Element(100,"fermium","Fm", 257, 0.2)
mendelevium =    Element(101,"mendelevium","Md", 258, 0.2)
nobelium =       Element(102,"nobelium","No", 259, 0.2)
lawrencium =     Element(103,"lawrencium","Lr", 262, 0.2)
rutherfordium =  Element(104,"rutherfordium","Rf", 261, 0.2)
dubnium =        Element(105,"dubnium","Db", 262, 0.2)
seaborgium =     Element(106,"seaborgium","Sg", 266, 0.2)
bohrium =        Element(107,"bohrium","Bh", 264, 0.2)
hassium =        Element(108,"hassium","Hs", 269, 0.2)
meitnerium =     Element(109,"meitnerium","Mt", 268, 0.2)
darmstadtium =   Element(110,"darmstadtium","Ds", 281, 0.2)
roentgenium =    Element(111,"roentgenium","Rg", 272, 0.2)
ununbium =       Element(112,"ununbium","Uub", 285, 0.2)
ununtrium =      Element(113,"ununtrium","Uut", 284, 0.2)
ununquadium =    Element(114,"ununquadium","Uuq", 289, 0.2)
ununpentium =    Element(115,"ununpentium","Uup", 288, 0.2)
ununhexium =     Element(116,"ununhexium","Uuh", 292, 0.2)

# Aliases to recognize common alternative spellings. Both the '==' and 'is'
# relational operators will work with any chosen name
sulphur = sulfur
aluminium = aluminum
virtual_site = virtual



# cython: c_string_type=str, c_string_encoding=ascii
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################

import os
import numpy as np
cimport numpy as np
np.import_array()
from mdtraj.utils import ensure_type, cast_indices, in_units_of
from mdtraj.utils.six import string_types
from mdtraj.formats.registry import FormatRegistry
from libc.stdlib cimport malloc, free
from libc.string cimport strcpy, strlen
from dcdlib cimport molfile_timestep_t, dcdhandle
from dcdlib cimport open_dcd_read, close_file_read, read_next_timestep
from dcdlib cimport open_dcd_write, close_file_write, write_timestep
from dcdlib cimport dcd_rewind


##############################################################################
# Globals
##############################################################################

__all__ = ['DCDTrajectoryFile', 'load_dcd']
# codes that indicate status on return from library
cdef int _DCD_SUCCESS    = 0   # No problems
cdef int _DCD_EOF    = -1   # No problems


cdef ERROR_MESSAGES = {
    -1: 'Normal EOF',
    -2: 'DCD file does not exist',
    -3: 'Open of DCD file failed',
    -4: 'Read call on DCD file failed',
    -5: 'Premature EOF found in DCD file',
    -6: 'Format of DCD file is wrong',
    -7: 'Output file already exists',
    -8: 'Malloc failed',
    -9: 'Write call on DCD file failed',
}

##############################################################################
# Code
##############################################################################

@FormatRegistry.register_loader('.dcd')
def load_dcd(filename, top=None, stride=None, atom_indices=None, frame=None):
    """load_dcd(filename, top=None, stride=None, atom_indices=None, frame=None)

    Load an DCD file from disk.

    The .dcd format is a cross-platform compressed binary trajectory format
    produced by many software packages, including CHARMM, NAMD, and OpenMM. It
    stores atomic coordinates, box vectors, and time information.

    Parameters
    ----------
    filename : path-like
        Path of DCD file.
    top : {path-like, Trajectory, Topology}
        DCD XTC format does not contain topology information. Pass in either
        the path to a pdb file, a trajectory, or a topology to supply this
        information.
    stride : int, default=None
        Only read every stride-th frame
    atom_indices : array_like, optional
        If not none, then read only a subset of the atoms coordinates from the
        file. This may be slightly slower than the standard read because it
        requires an extra copy, but will save memory.
    frame : int, optional
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.

    Examples
    --------
    >>> import mdtraj as md
    >>> traj = md.load_dcd('output.dcd', top='topology.pdb')
    >>> print traj
    <mdtraj.Trajectory with 500 frames, 423 atoms at 0x110740a90>

    >>> traj2 = md.load_dcd('output.dcd', stride=2, top='topology.pdb')
    >>> print traj2
    <mdtraj.Trajectory with 250 frames, 423 atoms at 0x11136e410>

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object.

    See Also
    --------
    mdtraj.DCDTrajectoryFile :  Low level interface to DCD files
    """
    from mdtraj.core.trajectory import _parse_topology

    # we make it not required in the signature, but required here. although this
    # is a little wierd, its good because this function is usually called by a
    # dispatch from load(), where top comes from **kwargs. So if its not supplied
    # we want to give the user an informative error message
    if top is None:
        raise ValueError('"top" argument is required for load_dcd')
    if not isinstance(filename, (string_types, os.PathLike)):
        raise TypeError('filename must be of type path-like for load_dcd. '
            'you supplied %s' % type(filename))

    topology = _parse_topology(top)
    atom_indices = cast_indices(atom_indices)

    with DCDTrajectoryFile(str(filename)) as f:
        if frame is not None:
            f.seek(frame)
            n_frames = 1
        else:
            n_frames = None
        return f.read_as_traj(topology, n_frames=n_frames, stride=stride, atom_indices=atom_indices)


cdef class DCDTrajectoryFile:
    """DCDTrajectoryFile(filename, mode='r', force_overwrite=True)

    Interface for reading and writing to a CHARMM/NAMD DCD file.
    This is a file-like object, that both reading or writing depending
    on the `mode` flag. It implements the context manager protocol,
    so you can also use it with the python 'with' statement.

    The conventional units in the DCD file are angstroms and degrees. The format
    supports saving coordinates and unit cell parameters (lengths and angles)

    Parameters
    ----------
    filename : string
        Path to the file to open
    mode : {'r', 'w'}
        Mode in which to open the file. 'r' is for reading, and 'w' is for writing.
    force_overwrite : bool
        In mode='w', how do you want to behave if a file by the name of `filename`
        already exists? if `force_overwrite=True`, it will be overwritten.

    Examples
    --------
    >>> # read a single frame, and then read the remaining frames
    >>> f = DCDTrajectoryFile('mytrajectory.dcd', 'r')
    >>> f.read(n_frames=1)  # read a single frame from the file
    >>> xyz.read()            # read all of the remaining frames
    >>> f.close()

    >>> # read all of the data with automatic closing of the file
    >>> with DCDTrajectoryFile('mytrajectory.dcd', 'r') as f:
    >>>    xyz, cell_lengths, cell_angles = f.read()

    >>> # write some xyz coordinates to a new file
    >>> with DCDTrajectoryFile('mytrajectory2.dcd', 'w') as f:
    >>>     f.write(np.random.randn(10,3,3))

    >>> # write frames one at a time
    >>> with DCDTrajectoryFile('mytrajectory2.dcd', 'w') as f:
    >>>     n_frames, n_atoms = 5, 10
    >>>     for i in range(n_frames):
    >>>         f.write(np.random.randn(n_atoms, 3))

    See Also
    --------
    mdtraj.load_dcd : High-level wrapper that returns a ``md.Trajectory``
    """

    # n_atoms and n_frames hold the number of atoms and the number of frames
    # in the file, as read off the header of the DCD file during read mode
    cdef int n_atoms, n_frames
    cdef dcdhandle* fh
    cdef char* mode
    cdef char* filename
    cdef int is_open, _needs_write_initialization
    cdef molfile_timestep_t* timestep
    # does this file have unitcell info? set during _initialize_write
    # for mode='w' to 0 or 1.
    cdef int with_unitcell
    cdef readonly char* distance_unit

    def __cinit__(self, char* filename, char* mode='r', force_overwrite=True):
        """Open a DCD Trajectory File
        """
        self.distance_unit = 'angstroms'
        self.is_open = False
        self.mode = mode

        # Note: we need to copy this string, see issue #206
        self.filename = <char*>malloc(strlen(filename)+1)
        if self.filename is NULL:
            raise MemoryError()
        strcpy(self.filename, filename)


        if str(mode) == 'r':
            self.fh = open_dcd_read(filename, "dcd", &self.n_atoms, &self.n_frames)
            if self.fh is NULL:
                raise IOError("Could not open file: %s" % filename)
            assert self.n_atoms > 0, 'DCD Corruption: n_atoms was not positive'
            assert self.n_frames >= 0, 'DCD corruption: n_frames < 0'
            self.is_open = True
        elif str(mode) == 'w':
            self._needs_write_initialization = 1
            if not force_overwrite and os.path.exists(filename):
                raise IOError('"%s" already exists' % filename)
        else:
            raise ValueError("most must be one of ['r', 'w']")

        # alloc the molfile_timestep, which is the struct that the library is
        # going to deposit its data into each timestep
        self.timestep = <molfile_timestep_t*> malloc(sizeof(molfile_timestep_t))
        if self.timestep is NULL:
            raise MemoryError('There was an error allocating memory')

    def __dealloc__(self):
        # free whatever we malloced
        free(self.timestep)
        free(self.filename)
        self.close()

    def _initialize_write(self, int n_atoms, int with_unitcell):
        """We don't actually want to open the dcd file during write mode
        until we know how many atoms the user wants to save. so this
        is delayed until the first call to write()
        """
        assert not self.is_open and self._needs_write_initialization

        self.n_atoms = n_atoms
        self.with_unitcell = with_unitcell
        self.fh = open_dcd_write(self.filename, "dcd", self.n_atoms, with_unitcell)
        if self.fh is NULL:
            raise IOError('There was an error opening the file: %s' % self.filename)
        self.is_open = True

        self._needs_write_initialization = False

    def close(self):
        """Close the DCD file handle
        """
        if self.is_open and self.fh is not NULL:
            if str(self.mode) == 'r':
                close_file_read(self.fh)
            else:
                close_file_write(self.fh)
            self.is_open = False

        self._needs_write_initialization = False

    def seek(self, offset, whence=0):
        """seek(offset, whence=0)

        Move to a new file position

        Parameters
        ----------
        offset : int
            A number of frames.
        whence : {0, 1, 2}
            0: offset from start of file, offset should be >=0.
            1: move relative to the current position, positive or negative
            2: move relative to the end of file, offset should be <= 0.
            Seeking beyond the end of a file is not supported
        """
        cdef int i, status
        if str(self.mode) != 'r':
            raise NotImplementedError("seek is only supported in mode='r'")

        advance, absolute = None, None
        current_pos = self.tell()
        if whence == 0 and offset >= 0:
            if offset >= current_pos:
                advance = offset - current_pos
            else:
                absolute = offset
        elif whence == 1 and offset >= 0:
            advance = offset
        elif whence == 1 and offset < 0:
            absolute = offset + current_pos
        elif whence == 2 and offset <= 0:
            raise NotImplementedError('offsets from the end are not supported yet')
        else:
            raise IOError('Invalid argument')

        if advance is not None:
            for i in range(advance):
                status = read_next_timestep(self.fh, self.n_atoms, NULL)
        elif absolute is not None:
            result = dcd_rewind(self.fh)
            if result != 0:
                raise IOError("Error seeking in %s" % self.filename)

            for i in range(absolute):
                status = read_next_timestep(self.fh, self.n_atoms, NULL)

    def tell(self):
        """Current file position

        Returns
        -------
        offset : int
            The current frame in the file.
        """
        return self.fh.setsread

    def __enter__(self):
        "Support the context manager protocol"
        return self

    def __exit__(self, *exc_info):
        "Support the context manager protocol"
        self.close()

    def __len__(self):
        """Number of frames in the file

        Notes
        -----
        This length is based on information in the header of the DCD
        file. It is possible for it to be off by 1 if, for instance,
        the client writing the file was killed between writing the last
        frame and updating the header information.
        """
        if not self.is_open:
            raise RuntimeError('I/O operation on closed file')
        return self.fh.nsets

    def read_as_traj(self, topology, n_frames=None, stride=None, atom_indices=None):
        """read_as_traj(topology, n_frames=None, stride=None, atom_indices=None)

        Read a trajectory from an XTC file

        Parameters
        ----------
        topology : Topology
            The system topology
        n_frames : int, optional
            If positive, then read only the next `n_frames` frames. Otherwise read all
            of the frames in the file.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        trajectory : Trajectory
            A trajectory object containing the loaded portion of the file.
        """
        from mdtraj.core.trajectory import Trajectory
        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        initial = self.tell()
        xyz, box_length, box_angle = self.read(n_frames=n_frames, stride=stride, atom_indices=atom_indices)
        if len(xyz) == 0:
            return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), topology=topology)

        in_units_of(xyz, self.distance_unit, Trajectory._distance_unit, inplace=True)
        in_units_of(box_length, self.distance_unit, Trajectory._distance_unit, inplace=True)

        if stride is None:
            stride = 1
        time = (stride*np.arange(len(xyz))) + initial

        return Trajectory(xyz=xyz, topology=topology, time=time,
                          unitcell_lengths=box_length,
                          unitcell_angles=box_angle)

    def read(self, n_frames=None, stride=None, atom_indices=None):
        """read(n_frames=None, stride=None, atom_indices=None)

        Read the data from a DCD file

        Parameters
        ----------
        n_frames : int, optional
            If positive, then read only the next `n_frames` frames. Otherwise read all
            of the frames in the file.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        xyz : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=float32
            The xyz coordinates of each atom in each frame. By convention, the
            coordinates in the dcd file are stored in units of angstroms.
        cell_lengths : np.ndarray, shape=(n_frames, 3), dtype=float32
            The length of the box in each frame. `cell_lengths[i,0]` is the length
            of the A axis (in frame i), and `cell_lengths[i,1]` and
            `cell_lengths[i,2]` are the B and C axis respectively. By convention,
            the cell lengths in the dcd file are stored in units of angstroms.
        cell_angles : np.ndarray, shape=(n_frames, 3), dtype=float32
            Organized analogously to cell_lengths. Gives the alpha, beta and
            gamma angles respectively in entries `cell_angles[i,0]`,
            `cell_angles[i,1]`, `cell_angles[i,2]`. By convention, the cell
            angles in the dcd file are stored in units of degrees.
        """
        if str(self.mode) != 'r':
            raise ValueError('read() is only available when the file is opened in mode="r"')
        if not self.is_open:
            raise IOError("file is not open")

        cdef int _n_frames, n_atoms_to_read, _stride
        if n_frames is None:
            # if the user specifies n_frames=None, they want to read to the
            # end of the file
            _n_frames = self.n_frames - self.tell()
        else:
            _n_frames = int(n_frames)

        if atom_indices is None:
            n_atoms_to_read = self.n_atoms
        elif isinstance(atom_indices, slice):
            n_atoms_to_read = len(np.arange(self.n_atoms)[atom_indices])
        else:
            if min(atom_indices) < 0:
                raise ValueError('atom_indices should be zero indexed. you gave an index less than zerp')
            if max(atom_indices) >= self.n_atoms:
                raise ValueError('atom indices should be zero indexed. you gave an index bigger than the number of atoms')
            n_atoms_to_read = len(atom_indices)

        if stride is None:
            _stride = 1
        else:
            _stride = stride

        # malloc space to put the data that we're going to read off the disk
        cdef np.ndarray[dtype=np.float32_t, ndim=3] xyz = np.zeros((_n_frames, n_atoms_to_read, 3), dtype=np.float32)
        cdef np.ndarray[dtype=np.float32_t, ndim=2] cell_lengths = np.zeros((_n_frames, 3), dtype=np.float32)
        cdef np.ndarray[dtype=np.float32_t, ndim=2] cell_angles = np.zeros((_n_frames, 3), dtype=np.float32)

        # only used if atom_indices is given
        cdef np.ndarray[dtype=np.float32_t, ndim=2] framebuffer = np.zeros((self.n_atoms, 3), dtype=np.float32)

        cdef int i, j
        cdef int status = _DCD_SUCCESS

        for i in range(_n_frames):
            if atom_indices is None:
                self.timestep.coords = &xyz[i,0,0]
                status = read_next_timestep(self.fh, self.n_atoms, self.timestep)
            else:
                self.timestep.coords = &framebuffer[0,0]
                status = read_next_timestep(self.fh, self.n_atoms, self.timestep)
                xyz[i, :, :] = framebuffer[atom_indices, :]

            cell_lengths[i, 0] = self.timestep.A
            cell_lengths[i, 1] = self.timestep.B
            cell_lengths[i, 2] = self.timestep.C
            cell_angles[i, 0] = self.timestep.alpha
            cell_angles[i, 1] = self.timestep.beta
            cell_angles[i, 2] = self.timestep.gamma

            if status != _DCD_SUCCESS:
                # if the frame was not successfully read, then we're done
                break

            for j in range(_stride - 1):
                if read_next_timestep(self.fh, self.n_atoms, NULL) != _DCD_SUCCESS:
                    break

        if np.all(cell_lengths < 1e-10):
            # in the DCD C code, if there's unitcell information inside the
            # DCD file, it just sets them to length=0 and angle=90. Other tools
            # like VMD explicitly write length=0 and angle=90 into their
            # DCD files. In both cases, we detect that indicate the
            # information's absense
            cell_lengths = None
            cell_angles = None

        if status == _DCD_SUCCESS:
            # if we're done either because of we read all of the n_frames
            # requested succcessfully, return
            return xyz, cell_lengths, cell_angles

        if status == _DCD_EOF:
            # if we're doing because we reached a normal EOF (perhaps the)
            # user asked to read more frames than were in the file, we need
            # to truncate the return arrays -- we don't want to return them
            # a big stack of zeros.
            # Note that a successful read prior an erroneous seek should not truncate,
            # this is why we do not assign the status of the seek operation.
            xyz = xyz[0:i]
            if cell_lengths is not None and cell_angles is not None:
                cell_lengths = cell_lengths[0:i]
                cell_angles = cell_angles[0:i]

            return xyz, cell_lengths, cell_angles

        # If we got some other status, thats a "real" error.
        raise IOError("Error: %s", ERROR_MESSAGES(status))

    def write(self, xyz, cell_lengths=None, cell_angles=None):
        """write(xyz, cell_lengths=None, cell_angles=None)

        Write one or more frames of data to the DCD file

        Parameters
        ----------
        xyz : np.ndarray, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of the atoms to write. By convention for
            this trajectory format, the lengths should be in units of angstroms.
        cell_lengths : np.ndarray, shape=(n_frames, 3), dtype=float32, optional
            The length of the periodic box in each frame, in each direction,
            `a`, `b`, `c`. By convention for
            this trajectory format, the lengths should be in units of angstroms.
        cell_angles : np.ndarray, shape=(n_frames, 3), dtype=float32, optional
            Organized analogously to cell_lengths. Gives the alpha, beta and
            gamma angles respectively. By convention for
            this trajectory format, the angles should be in units of degrees.
        """
        if str(self.mode) != 'w':
            raise ValueError('write() is only available when the file is opened in mode="w"')
        if not self._needs_write_initialization and not self.is_open:
            raise IOError("file is not open")

        # do typechecking, and then dispatch to the c level function
        xyz = ensure_type(xyz, dtype=np.float32, ndim=3, name='xyz', can_be_none=False,
                          add_newaxis_on_deficient_ndim=True)
        n_frames = len(xyz)

        # they must be both present or both absent
        if cell_lengths is None and cell_angles is not None:
            raise ValueError('cell_lengths were given, but no cell_angles')
        if cell_lengths is not None and cell_angles is None:
            raise ValueError('cell_angles were given, but no cell_lengths')

        cell_lengths = ensure_type(cell_lengths, dtype=np.float32, ndim=2, name='cell_lengths',
                                   can_be_none=True, shape=(n_frames, 3), add_newaxis_on_deficient_ndim=True,
                                   warn_on_cast=False)
        cell_angles = ensure_type(cell_angles, dtype=np.float32, ndim=2, name='cell_angles',
                                  can_be_none=True, shape=(n_frames, 3), add_newaxis_on_deficient_ndim=True,
                                  warn_on_cast=False)

        if self._needs_write_initialization:
            self._initialize_write(xyz.shape[1], (cell_lengths is not None) and (cell_angles is not None))
        else:
            if not self.n_atoms == xyz.shape[1]:
                raise ValueError('Number of atoms doesnt match')
            if cell_lengths is None and self.with_unitcell:
                raise ValueError("The file that you're saving to expects each frame "
                    "to contain unitcell information, but you did not supply it.")
            if cell_lengths is not None and not self.with_unitcell:
                raise ValueError("The file that you're saving to was created without "
                    "unitcell information.")

        self._write(xyz, cell_lengths, cell_angles)

    cdef _write(self, np.ndarray[np.float32_t, ndim=3, mode="c"] xyz,
                np.ndarray[np.float32_t, ndim=2, mode="c"] cell_lengths,
                np.ndarray[np.float32_t, ndim=2, mode="c"] cell_angles):
        if str(self.mode) != 'w':
            raise ValueError('_write() is only available when the file is opened in mode="w"')


        cdef int i, status
        cdef int n_frames = len(xyz)

        for i in range(n_frames):
            self.timestep.coords = &xyz[i, 0, 0]
            if cell_angles is not None and cell_lengths is not None:
                self.timestep.A = cell_lengths[i, 0]
                self.timestep.B = cell_lengths[i, 1]
                self.timestep.C = cell_lengths[i, 2]
                self.timestep.alpha = cell_angles[i, 0]
                self.timestep.beta  = cell_angles[i, 1]
                self.timestep.gamma = cell_angles[i, 2]
            # when the dcd handle is opened during initialize_write,
            # it passes the flag for whether the unitcell information is written
            # to disk or not. So when we don't have unitcell information being
            # written, the unitcell fields (A,B,C,alpha,beta,gamma) are ignored
            # during write_timestep()

            status = write_timestep(self.fh, self.timestep)

            if status != _DCD_SUCCESS:
                raise IOError("DCD Error: %s" % ERROR_MESSAGES(status))
FormatRegistry.register_fileobject('.dcd')(DCDTrajectoryFile)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon, Kyle Beauchamp
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################

import cython
import warnings
import numpy as np
from mdtraj.utils import ensure_type

from cpython cimport bool
from cython.parallel cimport prange

##############################################################################
# External Declarations
##############################################################################

cdef extern float msd_axis_major(int nrealatoms, int npaddedatoms, int rowstride,
    float* aT, float* bT, float G_a, float G_b) nogil
cdef extern float msd_atom_major(int nrealatoms, int npaddedatoms,  float* a,
    float* b, float G_a, float G_b, int computeRot, float rot[9]) nogil
cdef extern float rot_msd_atom_major(const int n_real_atoms,
    const int n_padded_atoms, const float* a, const float* b, const float rot[9]) nogil
cdef extern float rot_atom_major(const int n_atoms, float* a, const float rot[9]) nogil
cdef extern from "center.h":
    void inplace_center_and_trace_atom_major(float* coords, float* traces,
    const int n_frames, const int n_atoms) nogil
cdef extern from "math.h":
    float sqrtf(float x) nogil


##############################################################################
# External (Public) Functions
##############################################################################


@cython.boundscheck(False)
def rmsd(target, reference, int frame=0, atom_indices=None,
         ref_atom_indices=None, bool parallel=True, bool precentered=False):
    """rmsd(target, reference, frame=0, atom_indices=None, parallel=True, precentered=False)

    Compute RMSD of all conformations in target to a reference conformation.
    Note, this will center the conformations in place.

    Parameters
    ----------
    target : md.Trajectory
        For each conformation in this trajectory, compute the RMSD to
        a particular 'reference' conformation in another trajectory
        object.
    reference : md.Trajectory
        The object containing the reference conformation to measure distances
        to.
    frame : int, default=0
        The index of the conformation in `reference` to measure
        distances to.
    atom_indices : array_like, or None
        The indices of the atoms to use in the RMSD calculation. If not
        supplied, all atoms will be used.
    ref_atom_indices : array_like, or None
        Use these indices for the reference trajectory. If not supplied,
        the atom indices will be the same as those for target.
    parallel : bool
        Use OpenMP to calculate each of the RMSDs in parallel over
        multiple cores.
    precentered : bool, default=False
        Assume that the conformations are already centered at the origin, and that
        the "rmsd_traces" have been computed, as is done by
        `Trajectory.center_coordinates`. The "rmsd_traces" are intermediate
        calculations needed for the RMSD calculation which can be computed
        independently on each trajectory. Note that this has the potential to
        be unsafe; if you use Trajectory.center_coordinates and then modify
        the trajectory's coordinates, the center and traces will be out of
        date and the RMSDs will be incorrect.

    Examples
    --------
    >>> import mdtraj as md                                      # doctest: +SKIP
    >>> rmsds = md.rmsd(trajectory, trajectory, 0)               # doctest: +SKIP
    >>> print rmsds                                              # doctest: +SKIP
    array([ 0.0,  0.03076187,  0.02549562, ...,  0.06230228,
        0.00666826,  0.24364147])

    The calculation is slightly faster if you precenter the trajectory

    >>> trajectory.center_coordinates()
    >>> rmsds = md.rmsd(trajectory, trajectory, 0, precentered=True)

    See Also
    --------
    Trajectory.center_coordinates

    Notes
    -----
    This function uses OpenMP to parallelize the calculation across
    multiple cores. To control the number of threads launched by OpenMP,
    you can set the environment variable ``OMP_NUM_THREADS``.

    Returns
    -------
    rmsds : np.ndarray, shape=(target.n_frames,)
        A 1-D numpy array of the optimal root-mean-square deviations from
        the `frame`-th conformation in reference to each of the conformations
        in target.
    """
    # import time
    cdef bool atom_indices_is_none = False

    if atom_indices is None:
        atom_indices_is_none = True
        atom_indices = slice(None)
    else:
        atom_indices = ensure_type(np.asarray(atom_indices), dtype=int, ndim=1, name='atom_indices')
        if not np.all((atom_indices >= 0) *
                              (atom_indices < target.xyz.shape[1])):
            raise ValueError("atom_indices must be valid positive indices")

    if ref_atom_indices is None:
        ref_atom_indices = atom_indices
    else:
        if len(ref_atom_indices) != len(atom_indices):
            raise ValueError("atom_indices and ref_atom_indices must have same number of atom indices. "
                             "found %d and %d." % (len(atom_indices), len(ref_atom_indices)))

    if not isinstance(ref_atom_indices, slice):
        ref_atom_indices = ensure_type(np.asarray(ref_atom_indices), dtype=int, ndim=1, name='ref_atom_indices')
        if not np.all((ref_atom_indices >= 0) *
                              (ref_atom_indices < reference.xyz.shape[1])):
            raise ValueError("ref_atom_indices must be valid positive indices")

    # Error checks
    assert (target.xyz.ndim == 3) and (reference.xyz.ndim == 3) and (target.xyz.shape[2]) == 3 and (reference.xyz.shape[2] == 3)
    if not ((target.xyz.shape[1]  == reference.xyz.shape[1]) or
            (len(ref_atom_indices) == len(atom_indices))):
        raise ValueError("Input trajectories must have same number of atoms. "
                         "found %d and %d." % (target.xyz.shape[1], reference.xyz.shape[1]))
    if frame >= reference.xyz.shape[0]:
        raise ValueError("Cannot calculate RMSD of frame %d: reference has "
                         "only %d frames." % (frame, reference.xyz.shape[0]))

    # static declarations
    cdef int i
    cdef float msd, ref_g
    cdef float[:, :, :] target_xyz
    cdef float[:, :] ref_xyz_frame
    cdef float[:] target_g
    cdef int target_n_frames = target.xyz.shape[0]
    cdef int n_atoms = target.xyz.shape[1] if np.all(atom_indices == slice(None)) else len(atom_indices)

    # make sure *every* frame in target_xyz is in proper c-major order
    target_xyz = np.asarray(target.xyz[:, atom_indices, :], order='C', dtype=np.float32)
    # only extract the `frame`-th conformation from ref_xyz
    ref_xyz_frame = np.asarray(reference.xyz[frame, ref_atom_indices, :], order='C', dtype=np.float32)

    # t0 = time.time()
    if precentered and (reference._rmsd_traces is not None) and (target._rmsd_traces is not None) and atom_indices_is_none:
        target_g = np.asarray(target._rmsd_traces, order='C', dtype=np.float32)
        ref_g = reference._rmsd_traces[frame]
    else:
        if precentered:
            warnings.warn(
                'in rmsd(), precentered is ignored when atom_indices != None',
                RuntimeWarning)
        target_g = np.empty(target_n_frames, dtype=np.float32)
        inplace_center_and_trace_atom_major(&target_xyz[0,0,0], &target_g[0], target_n_frames, n_atoms)
        inplace_center_and_trace_atom_major(&ref_xyz_frame[0, 0], &ref_g, 1, n_atoms)

    # t1 = time.time()

    cdef float[:] distances = np.zeros(target_n_frames, dtype=np.float32)
    if parallel:
        for i in prange(target_n_frames, nogil=True):
            msd = msd_atom_major(n_atoms, n_atoms, &target_xyz[i, 0, 0], &ref_xyz_frame[0, 0], target_g[i], ref_g, 0, NULL)
            distances[i] = sqrtf(msd)
    else:
        for i in range(target_n_frames):
            msd = msd_atom_major(n_atoms, n_atoms, &target_xyz[i, 0, 0], &ref_xyz_frame[0, 0], target_g[i], ref_g, 0, NULL)
            distances[i] = sqrtf(msd)

    # t2 = time.time()
    # print 'rmsd: %s, centering: %s' % (t2-t1, t1-t0)
    return np.array(distances, copy=False)


@cython.boundscheck(False)
def rmsf(target, reference, int frame=0, atom_indices=None,
         ref_atom_indices=None, bool parallel=True, bool precentered=False):
    """rmsf(target, reference, frame=0, atom_indices=None, parallel=True, precentered=False)
    
    Compute RMSF of atom positions in target trajectory. This will center target conformations in place.
    
    Parameters
    ----------
    target : md.Trajectory
        Compute the RMSF of atom positions in target trajectory. 
    reference : md.Trajectory, or None
        A trajectory with the same number of atoms as in the target to be used
        as a reference. If None, the average xyz positions of each atom will be used.
    frame : int, default=0
        The index of the conformation in `reference` to measure
        distances to.
    atom_indices : array_like, or None
        The indices of the atoms to use in the RMSF calculation. If not
        supplied, all atoms will be used. If the atoms used in alignment
        and the atoms used in RMSF calculations are different, the trajectory
        must be aligned first manually.
    ref_atom_indices : array_like, or None
        Use these indices for the reference trajectory. If not supplied,
        the atom indices will be the same as those for target.
    parallel : bool
        Use OpenMP to calculate each of the RMSFs in parallel over
        multiple cores.
    precentered : bool, default=False
        Assume that the conformations are already centered at the origin, and that
        the "rmsd_traces" have been computed, as is done by
        `Trajectory.center_coordinates`. The "rmsd_traces" are intermediate
        calculations needed for the RMSF calculation which can be computed
        independently on each trajectory. Note that this has the potential to
        be unsafe; if you use Trajectory.center_coordinates and then modify
        the trajectory's coordinates, the center and traces will be out of
        date and the RMSFs will be incorrect.
    
    Examples
    --------
    >>> import mdtraj as md                                      # doctest: +SKIP
    >>> rmsf = md.rmsf(trajectory, trajectory, 0)                # doctest: +SKIP
    >>> print rmsf                                               # doctest: +SKIP
    array([ 0.0,  0.03076187,  0.02549562, ...,  0.06230228,
        0.00666826,  0.24364147])
    
    The calculation is slightly faster if you precenter the trajectory
    >>> trajectory.center_coordinates()
    >>> rmsf = md.rmsf(trajectory, trajectory, 0, precentered=True)
    
    If the atoms used in alignment and RMSF calculations are different,
    align the trajectory before using this method.
    >>> trajectory.superpose(atom_indices=atom_indices)
    >>> rmsf = md.rmsf(trajectory, None, atom_indices=rmsf_atom_indices)
    
    See Also
    --------
    Trajectory.center_coordinates
    
    Notes
    -----
    This function uses OpenMP to parallelize the calculation across
    multiple cores. To control the number of threads launched by OpenMP,
    you can set the environment variable ``OMP_NUM_THREADS``.
    
    Returns
    -------
    rmsf : np.ndarray, shape=(atom_indices,)
        A 1-D numpy array of the optimal root-mean-square fluctuations of the
       selected atoms.
    """
    # import time
    cdef bool atom_indices_is_none = False
    cdef bool trajectory_prealigned = False

    if reference is None:
        trajectory_prealigned = True
        reference = target

    if atom_indices is None:
        atom_indices_is_none = True
        atom_indices = slice(None)
    else:
        atom_indices = ensure_type(np.asarray(atom_indices), dtype=int, ndim=1, name='atom_indices')
        if not np.all((atom_indices >= 0) *
                              (atom_indices < target.xyz.shape[1]) *
                              (atom_indices < reference.xyz.shape[1])):
            raise ValueError("atom_indices must be valid positive indices")

    if ref_atom_indices is None:
        ref_atom_indices = atom_indices
    else:
        if len(ref_atom_indices) != len(atom_indices):
            raise ValueError("atom_indices and ref_atom_indices must have same number of atom indices. "
                             "found %d and %d." % (len(atom_indices), len(ref_atom_indices)))

    if not isinstance(ref_atom_indices, slice):
        ref_atom_indices = ensure_type(np.asarray(ref_atom_indices), dtype=int, ndim=1, name='ref_atom_indices')
        if not np.all((ref_atom_indices >= 0) *
                              (ref_atom_indices < target.xyz.shape[1]) *
                              (ref_atom_indices < reference.xyz.shape[1])):
            raise ValueError("ref_atom_indices must be valid positive indices")

    # Error checks
    assert (target.xyz.ndim == 3) and (reference.xyz.ndim == 3) and (target.xyz.shape[2]) == 3 and (reference.xyz.shape[2] == 3)
    if not ((target.xyz.shape[1]  == reference.xyz.shape[1]) or
            (len(ref_atom_indices) == len(atom_indices))):
        raise ValueError("Input trajectories must have same number of atoms. "
                         "found %d and %d." % (target.xyz.shape[1], reference.xyz.shape[1]))
    if frame >= reference.xyz.shape[0]:
        raise ValueError("Cannot calculate RMSF of frame %d: reference has "
                         "only %d frames." % (frame, reference.xyz.shape[0]))

    # static declarations
    cdef int i, j
    cdef float msd, ref_g
    cdef float[:, :, :] target_xyz
    cdef float[:, :] ref_xyz_frame
    cdef float[:, :, :] target_displaced_xyz
    cdef float[:, :] avg_xyz_frame
    cdef float[:] target_g
    cdef int target_n_frames = target.xyz.shape[0]
    cdef int n_atoms = target.xyz.shape[1] if np.all(atom_indices == slice(None)) else len(atom_indices)

    # make sure *every* frame in target_xyz is in proper c-major order
    target_xyz = np.asarray(target.xyz[:, atom_indices, :], order='C', dtype=np.float32)
    # only extract the `frame`-th conformation from ref_xyz
    ref_xyz_frame = np.asarray(reference.xyz[frame, ref_atom_indices, :], order='C', dtype=np.float32)
    avg_xyz_frame = np.zeros_like(reference.xyz[frame, ref_atom_indices, :], order='C', dtype=np.float32)

    # t0 = time.time()
    if precentered and (reference._rmsd_traces is not None) and (target._rmsd_traces is not None) and atom_indices_is_none:
        target_g = np.asarray(target._rmsd_traces, order='C', dtype=np.float32)
        ref_g = reference._rmsd_traces[frame]
    else:
        if precentered:
            warnings.warn(
                'in rmsd(), precentered is ignored when atom_indices != None',
                RuntimeWarning)
        target_g = np.empty(target_n_frames, dtype=np.float32)
        inplace_center_and_trace_atom_major(&target_xyz[0,0,0], &target_g[0], target_n_frames, n_atoms)
        inplace_center_and_trace_atom_major(&ref_xyz_frame[0, 0], &ref_g, 1, n_atoms)

    # t1 = time.time()

    cdef float[:] fluctuations = np.zeros(n_atoms, dtype=np.float32)
    # we could get away with using less memory here: we only need 1 rotation
    # matrix in serial mode and 1 rotation matrix per thread in parallel mode.
    cdef float[:, :, ::1] rot = np.zeros((target_n_frames, 3, 3), dtype=np.float32)
    target_displaced_xyz = np.array(target.xyz[:, atom_indices, :], copy=True)

    if trajectory_prealigned == False:
        if parallel:
            for i in prange(target_n_frames, nogil=True):
                msd_atom_major(n_atoms, n_atoms, &target_xyz[i, 0, 0], &ref_xyz_frame[0, 0], ref_g, target_g[i], 1, &rot[i, 0, 0])
                rot_atom_major(n_atoms, &target_displaced_xyz[i, 0, 0], &rot[i, 0, 0])
        else:
            for i in range(target_n_frames):
                msd_atom_major(n_atoms, n_atoms, &target_xyz[i, 0, 0], &ref_xyz_frame[0, 0], ref_g, target_g[i], 1, &rot[i, 0, 0])
                rot_atom_major(n_atoms, &target_displaced_xyz[i, 0, 0], &rot[i, 0, 0])

    if parallel:
        for j in prange(n_atoms, nogil=True):
            for i in range(target_n_frames):
                avg_xyz_frame[j, 0] += target_displaced_xyz[i, j, 0] / target_n_frames
                avg_xyz_frame[j, 1] += target_displaced_xyz[i, j, 1] / target_n_frames
                avg_xyz_frame[j, 2] += target_displaced_xyz[i, j, 2] / target_n_frames
    else:
        for i in range(target_n_frames):
            for j in range(n_atoms):
                avg_xyz_frame[j, 0] += target_displaced_xyz[i, j, 0] / target_n_frames
                avg_xyz_frame[j, 1] += target_displaced_xyz[i, j, 1] / target_n_frames
                avg_xyz_frame[j, 2] += target_displaced_xyz[i, j, 2] / target_n_frames

    if parallel:
        for j in prange(n_atoms, nogil=True):
            for i in range(target_n_frames):
                fluctuations[j] += ((target_displaced_xyz[i, j, 0] - avg_xyz_frame[j, 0])**2 +
                                    (target_displaced_xyz[i, j, 1] - avg_xyz_frame[j, 1])**2 +
                                    (target_displaced_xyz[i, j, 2] - avg_xyz_frame[j, 2])**2) / target_n_frames
    else:
        for i in range(target_n_frames):
            for j in range(n_atoms):
                fluctuations[j] += ((target_displaced_xyz[i, j, 0] - avg_xyz_frame[j, 0])**2 +
                                    (target_displaced_xyz[i, j, 1] - avg_xyz_frame[j, 1])**2 +
                                    (target_displaced_xyz[i, j, 2] - avg_xyz_frame[j, 2])**2) / target_n_frames

    for j in range(n_atoms):
        fluctuations[j] = sqrtf(fluctuations[j])

    # t2 = time.time()
    # print 'rmsd: %s, centering: %s' % (t2-t1, t1-t0)
    return np.array(fluctuations, copy=False)


def _center_inplace_atom_major(float[:, :, ::1] xyz not None):
    assert xyz.shape[2] == 3
    cdef float[:] traces = np.empty(xyz.shape[0], dtype=np.float32)
    inplace_center_and_trace_atom_major(&xyz[0,0,0], &traces[0], xyz.shape[0], xyz.shape[1])
    return np.array(traces, copy=False)



##############################################################################
# Private Functions
##############################################################################


@cython.boundscheck(False)
@cython.wraparound(False)
def getMultipleRMSDs_axis_major(float[:, :, ::1] xyz1 not None,
                                float[:, :, ::1] xyz2 not None, float[::1] g1,
                                float[::1] g2, int frame, bool parallel=True):
    """getMultipleRMSDs_axis_major(xyz1, xyz2, g1, g, frame, parallel=True)

    Calculate the RMSD of several frames to a single frame, with the
    coordinates laid out in axis-major orders

    Parameters
    ----------
    xyz1 : np.ndarray, shape=(n_frames, 3, n_atoms), dtype=float32
        Coordinates of reference frame.
    xyz2 : np.ndarray, shape=(n_frames, 3, n_atoms), dtype=float32
        Coordinates of frames to iterate over
    g1 : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz1
    g2 : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz2
    frame : int
        Index of the desired reference frame in xyz1.
    parallel : bool, default True
        If True, use OpenMP parallelization.

    Returns
    -------
    rmsds: np.ndarray, shape=(len(xyz2),)
        RMSDS between xyz1[frame] and all of xyz2
    """

    cdef Py_ssize_t i
    cdef int n_frames = xyz2.shape[0]
    cdef int n_atoms = xyz1.shape[2]
    cdef float msd

    assert xyz1.ndim == 3 and xyz2.ndim == 3 and xyz1.shape[1] == 3 and xyz2.shape[1] == 3
    if not (xyz1.shape[2]  == xyz2.shape[2]):
        raise ValueError("Input arrays must have same number of atoms. "
                         "found %d and %d." % (xyz1.shape[2], xyz2.shape[2]))
    if frame >= xyz1.shape[0]:
        raise ValueError("Cannot calculate RMSD of frame %d: xyz1 has "
                         "only %d frames." % (frame, xyz1.shape[0]))

    cdef float[:] distances = np.zeros(n_frames, dtype=np.float32)

    if parallel == True:
        for i in prange(n_frames, nogil=True):
            msd = msd_axis_major(n_atoms, n_atoms, n_atoms,
                                   &xyz1[frame, 0, 0], &xyz2[i, 0, 0], g1[frame], g2[i])
            distances[i] = sqrtf(msd)
    else:
        for i in range(n_frames):
            msd = msd_axis_major(n_atoms, n_atoms, n_atoms,
                                   &xyz1[frame, 0, 0], &xyz2[i, 0, 0], g1[frame], g2[i])
            distances[i] = sqrtf(msd)

    return np.array(distances, copy=False)


@cython.boundscheck(False)
@cython.wraparound(False)
def getMultipleRMSDs_atom_major(float[:, :, ::1] xyz1 not None, float[:, :, ::1] xyz2 not None,
                                float[::1] g1 not None, float[::1] g2 not None, int frame,
                                bool parallel=True):
    """getMultipleRMSDs_atom_major(xyz1, xyz2, g1, g2, n_atoms, frame, parallel=True)

    Calculate the RMSD of several frames to a single frame, with the
    coordinates laid out in atom-major orders

    Parameters
    ----------
    xyz1 : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=float32
        Coordinates of reference frame.
    xyz2 : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=float32
        Coordinates of frames to iterate over
    g1 : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz1
    g2 : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz2
    frame : int
        Index of the desired reference frame in xyz1.
    parallel : bool, default True
        If True, use OpenMP parallelization.

    Returns
    -------
    rmsds: np.ndarray, shape=(len(xyz2),)
        RMSDS between xyz1[frame] and all of xyz2
    """

    cdef Py_ssize_t i
    cdef int n_frames = xyz2.shape[0]
    cdef int n_atoms = xyz1.shape[1]
    cdef float msd

    assert xyz1.ndim == 3 and xyz2.ndim == 3 and xyz1.shape[2] == 3 and xyz2.shape[2] == 3
    if not (xyz1.shape[1]  == xyz2.shape[1]):
        raise ValueError("Input arrays must have same number of atoms. "
                         "found %d and %d." % (xyz1.shape[1], xyz2.shape[1]))
    if frame >= xyz1.shape[0]:
        raise ValueError("Cannot calculate RMSD of frame %d: xyz1 has "
                         "only %d frames." % (frame, xyz1.shape[0]))

    cdef float[:] distances = np.zeros(n_frames, dtype=np.float32)

    if parallel == True:
        for i in prange(n_frames, nogil=True):
            msd = msd_atom_major(n_atoms, n_atoms,  &xyz1[frame, 0, 0], &xyz2[i, 0, 0], g1[frame], g2[i], 0, NULL)
            distances[i] = sqrtf(msd)
    else:
        for i in range(n_frames):
            msd = msd_atom_major(n_atoms, n_atoms, &xyz1[frame, 0, 0], &xyz2[i, 0, 0], g1[frame], g2[i], 0, NULL)
            distances[i] = sqrtf(msd)

    return np.array(distances, copy=False)


@cython.boundscheck(False)
@cython.wraparound(False)
def superpose_atom_major(float[:, :, ::1] xyz_align_target not None,
                         float[:, :, ::1] xyz_align_mobile not None,
                         float[::1] g_target not None, float[::1] g_mobile not None,
                         float[:, :, ::1] xyz_displace_mobile not None, int target_frame,
                         bool parallel=True):
    """superpose_atom_major(xyz_target, xyz_mobile, g_target, g_mobile, xyz_mobile_displace)

    Superpose each frame in xyz2 upon a frame in xyz1

    Parameters
    ----------
    xyz_align_target : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=float32
        Coordinates of reference frame to align to.
    xyz_align_mobile: np.ndarray, shape=(n_frames, n_atoms, 3), dtype=float32
        Coordinates of the mobile trajectory to align to target the target.
    g_target : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz_target
    g_mobile : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz_mobile
    xyz_displace_mobile : np.ndarray, shape=(n_frames, n_atoms, 3), dtype=float32
        The coordinates of the mobile trajectory to displace
    target_frame : int
        The particular frame in xyz_align_target / g_target to align the mobile
        trajectory  to.
    parallel : bool, default=True
        Run the calculation using multiple cores simultaneously.
    """
    cdef int i
    cdef int n_frames = xyz_align_mobile.shape[0]
    cdef int n_atoms_align = xyz_align_target.shape[1]
    cdef int n_atoms_displace = xyz_displace_mobile.shape[1]
    if not (xyz_align_target.shape[1]  == xyz_align_mobile.shape[1]):
        raise ValueError("Input arrays must have same number of atoms. "
                         "found %d and %d." % (
                         xyz_align_target.shape[1],
                         xyz_align_mobile.shape[1]))
    if not xyz_align_mobile.shape[0] == xyz_displace_mobile.shape[0]:
        raise ValueError("xyz_align_mobile and xyz_displace_mobile must contain the same number of frames")

    # we could get away with using less memory here: we only need 1 rotation
    # matrix in serial mode and 1 rotation matrix per thread in parallel mode.
    cdef float[:, :, ::1] rot = np.zeros((n_frames, 3, 3), dtype=np.float32)

    if parallel == True:
        for i in prange(n_frames, nogil=True):
            msd_atom_major(n_atoms_align, n_atoms_align, &xyz_align_mobile[i, 0, 0],
                           &xyz_align_target[target_frame, 0, 0],
                           g_target[target_frame], g_mobile[i], 1, &rot[i, 0, 0])
            rot_atom_major(n_atoms_displace, &xyz_displace_mobile[i, 0, 0], &rot[i, 0, 0])
    else:
        for i in range(n_frames):
            msd_atom_major(n_atoms_align, n_atoms_align, &xyz_align_mobile[i, 0, 0],
                           &xyz_align_target[target_frame, 0, 0],
                           g_target[target_frame], g_mobile[i], 1, &rot[i, 0, 0])
            rot_atom_major(n_atoms_displace, &xyz_displace_mobile[i, 0, 0], &rot[i, 0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
def getMultipleAlignDisplaceRMSDs_atom_major(float[:, :, ::1] xyz_align1 not None,
    float[:, :, ::1] xyz_align2 not None, float[::1] g_align1 not None,
    float[::1] g_align2 not None, float[:, :, ::1] xyz_displ1 not None,
    float[:, :, ::1] xyz_displ2 not None, int n_atoms_align,
    int n_atoms_displ, int frame, bool parallel=True):
    """getMultipleAlignDisplaceRMSDs_atom_major(xyz_align1, xyz_align2, g_align1, g_align2, xyz_displ1, xyz_displ2, n_atoms_align, n_atoms_displ, frame, parallel=True)

    Calculate the RMSD of several frames to a single frame, with the
    coordinates laid out in atom-major orders, where alignments are computed
    by determining the optimal rotation matrix to align `xyz_align1` against
    `xyz_align2`, and then using this rotation matrix displace `xyz_displ1`,
    and computing its mean-squared-displacement with respect to `xyz_displ2`.

    Parameters
    ----------
    xyz_align1 : np.ndarray, shape=(n_frames, n_atoms_align_padded, 3), dtype=float32
        Alignment coordinates of reference frame.
    xyz_align2 : np.ndarray, shape=(n_frames, n_atoms_align_padded, 3), dtype=float32
        Alignment coordinates of frames to iterate over.
    g_align1 : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz1
    g_align2 : np.ndarray, shape = (n_frames), dtype=float32
        Pre-calculated G factors (traces) for each frame in xyz2
    xyz_displ1 : np.ndarray, shape=(n_frames, n_atoms_displ_padded, 3), dtype=float32
        Displacement coordinates of reference frame.
    xyz_displ2 : np.ndarray, shape=(n_frames, n_atoms_displ_padded, 3), dtype=float32
        Displacement coordinates of frames to iterate over.
    n_atoms_align : int
        The number of true atoms in the alignment coordinates
    n_atoms_displ : int
        The number of true atoms in the displacement coordinates
    frame : int
        Index of the desired reference frame in xyz_align1 / xyz_displ1.
    parallel : bool, default True
        If True, use OpenMP parallelization.

    Returns
    -------
    rmsds: np.ndarray, shape=(len(xyz2),)
        RMSDs between xyz_displ1[frame] and all of xyz_displ2, after the alignment that maximizes
        the overlap between xyz_align1[frame] and all of xyz_align2
    rotations: np.ndarray, shape=(len(xyz2), 3, 3)
        The rotation matrix for each of the alignments
    """
    assert (xyz_align1.shape[2] == 3) and (xyz_align2.shape[2] == 3)
    assert (xyz_displ1.shape[2] == 3) and (xyz_displ2.shape[2] == 3)
    if not ((xyz_align1.shape[1] % 4 == 0) & (xyz_align2.shape[1] % 4 == 0)):
        raise ValueError("Input arrays must have middle dimension of 4*n, "
                         "found %d and %d." % (xyz_align1.shape[1], xyz_align2.shape[1]))
    if not ((xyz_displ1.shape[1] % 4 == 0) & (xyz_displ2.shape[1] % 4 == 0)):
        raise ValueError("Input arrays must have middle dimension of 4*n, "
                         "found %d and %d." % (xyz_displ1.shape[1], xyz_displ2.shape[1]))
    if not xyz_align1.shape[0] == xyz_displ1.shape[0]:
        raise ValueError("xyz_align1 and xyz_displ1 must contain the same number of frames")
    if not xyz_align2.shape[0] == xyz_displ2.shape[0]:
        raise ValueError("xyz_align2 and xyz_displ2 must contain the same number of frames")
    if frame >= xyz_align1.shape[0]:
        raise ValueError("Cannot calculate RMSD of frame %d: xyz1 has "
                         "only %d frames." % (frame, xyz_align1.shape[0]))

    cdef float msd
    cdef Py_ssize_t i
    cdef int n_frames = xyz_align2.shape[0]
    cdef int n_align_atoms_padded = xyz_align1.shape[1]
    cdef int n_displ_atoms_padded = xyz_displ1.shape[1]

    cdef float[:, :, ::1] rot = np.zeros((n_frames, 3, 3), dtype=np.float32)
    cdef float[:] distances = np.zeros(n_frames, dtype=np.float32)

    if parallel == True:
        for i in prange(n_frames, nogil=True):
            msd_atom_major(n_atoms_align, n_align_atoms_padded, &xyz_align1[frame, 0, 0], &xyz_align2[i, 0, 0], g_align1[frame], g_align2[i], 1, &rot[i, 0, 0])
            msd = rot_msd_atom_major(n_atoms_displ, n_displ_atoms_padded, &xyz_displ1[frame, 0, 0], &xyz_displ2[i, 0, 0], &rot[i, 0, 0])
            distances[i] = sqrtf(msd)
    else:
        for i in range(n_frames):
            msd_atom_major(n_atoms_align, n_align_atoms_padded, &xyz_align1[frame, 0, 0], &xyz_align2[i, 0, 0], g_align1[frame], g_align2[i], 1, &rot[i, 0, 0])
            msd = rot_msd_atom_major(n_atoms_displ, n_displ_atoms_padded, &xyz_displ1[frame, 0, 0], &xyz_displ2[i, 0, 0], &rot[i, 0, 0])
            distances[i] = sqrtf(msd)

    return np.array(distances, copy=False), np.array(rot, copy=False)




from __future__ import print_function, division
import time
import warnings
from mdtraj.utils.delay_import import import_
from mdtraj.utils.validation import ensure_type, cast_indices, check_random_state
from mdtraj.utils.unit import in_units_of
from mdtraj.utils.rotation import rotation_matrix_from_quaternion, uniform_quaternion
from mdtraj.utils.unitcell import (lengths_and_angles_to_box_vectors,
                       box_vectors_to_lengths_and_angles,
                       lengths_and_angles_to_tilt_factors)
from mdtraj.utils.contextmanagers import timing, enter_temp_directory
from mdtraj.utils.zipped import open_maybe_zipped

__all__ = ["ensure_type", "import_", "in_units_of",
           "lengths_and_angles_to_box_vectors",
           "box_vectors_to_lengths_and_angles",
           "tilt_factors_to_angles",
           "ilen", "timing", "cast_indices", "check_random_state",
           "rotation_matrix_from_quaternion", "uniform_quaternion",
           "enter_temp_directory", "timing", "deprecated"]


def ilen(iterable):
    """Length of an iterator. Note, this consumes the iterator

    Parameters
    ----------
    iterable : iterable
        An iterable, such as a generator, list, etc.

    Returns
    -------
    length  : int
        The number of elements in the iterable
    """
    return sum(1 for _ in iterable)


class deprecated(object):
    """Decorator to mark a function or class as deprecated.

    Issue a warning when the function is called/the class is instantiated and
    adds a warning to the docstring.

    The optional extra argument will be appended to the deprecation message
    and the docstring. Note: to use this with the default value for extra, put
    in an empty of parentheses:

    >>> from sklearn.utils import deprecated
    >>> deprecated() # doctest: +ELLIPSIS
    <sklearn.utils.deprecated object at ...>

    >>> @deprecated()
    ... def some_function(): pass
    """
    # Copied from scikit-learn: sklearn/utils/__init__.py
    # Adapted from http://wiki.python.org/moin/PythonDecoratorLibrary,
    # but with many changes.

    def __init__(self, extra=''):
        """
        Parameters
        ----------
        extra: string
          to be added to the deprecation messages

        """
        self.extra = extra

    def __call__(self, obj):
        if isinstance(obj, type):
            return self._decorate_class(obj)
        else:
            return self._decorate_fun(obj)

    def _decorate_class(self, cls):
        msg = "Class %s is deprecated" % cls.__name__
        if self.extra:
            msg += "; %s" % self.extra

        # FIXME: we should probably reset __new__ for full generality
        init = cls.__init__

        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return init(*args, **kwargs)
        cls.__init__ = wrapped

        wrapped.__name__ = '__init__'
        wrapped.__doc__ = self._update_doc(init.__doc__)
        wrapped.deprecated_original = init

        return cls

    def _decorate_fun(self, fun):
        """Decorate function fun"""

        msg = "Function %s is deprecated" % fun.__name__
        if self.extra:
            msg += "; %s" % self.extra

        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return fun(*args, **kwargs)

        wrapped.__name__ = fun.__name__
        wrapped.__dict__ = fun.__dict__
        wrapped.__doc__ = self._update_doc(fun.__doc__)

        return wrapped

    def _update_doc(self, olddoc):
        newdoc = "DEPRECATED"
        if self.extra:
            newdoc = "%s: %s" % (newdoc, self.extra)
        if olddoc:
            newdoc = "%s\n\n%s" % (newdoc, olddoc)
        return newdoc



"""MDTraj: A modern, open library for the analysis of molecular dynamics trajectories

MDTraj is a python library that allows users to manipulate molecular dynamics
(MD) trajectories and perform a variety of analyses, including fast RMSD,
solvent accessible surface area, hydrogen bonding, etc. A highlight of MDTraj
is the wide variety of molecular dynamics trajectory file formats which are
supported, including RCSB pdb, GROMACS xtc, tng, and trr, CHARMM / NAMD dcd, AMBER
binpos, AMBER NetCDF, AMBER mdcrd, TINKER arc and MDTraj HDF5.
"""
from __future__ import print_function, absolute_import

import sys
from glob import glob

DOCLINES = __doc__.split("\n")

from setuptools import setup, Extension, find_packages

sys.path.insert(0, '.')
from basesetup import (write_version_py, build_ext,
                       StaticLibrary, CompilerDetection, parse_setuppy_commands)

try:
    # add an optional --disable-openmp to disable OpenMP support
    sys.argv.remove('--disable-openmp')
    disable_openmp = True
except ValueError:
    disable_openmp = False


##########################
VERSION = "1.9.7"
ISRELEASED = True
__version__ = VERSION
##########################


CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)
Programming Language :: C
Programming Language :: Python
Programming Language :: Python :: 3
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Chemistry
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

# Global info about compiler
compiler = CompilerDetection(disable_openmp)
compiler.initialize()

extra_cpp_libraries = []

if sys.platform == 'win32':
    extra_cpp_libraries.append('Ws2_32')
    # For determining if a path is relative (for dtr)
    extra_cpp_libraries.append('Shlwapi')


################################################################################
# Declaration of the compiled extension modules (cython + c)
################################################################################

def format_extensions():
    compiler_args = compiler.compiler_args_warn

    xtc = Extension('mdtraj.formats.xtc',
                    sources=['mdtraj/formats/xtc/src/xdrfile.c',
                             'mdtraj/formats/xtc/src/xdr_seek.c',
                             'mdtraj/formats/xtc/src/xdrfile_xtc.c',
                             'mdtraj/formats/xtc/xtc.pyx',
                             ],
                    include_dirs=['mdtraj/formats/xtc/include/',
                                  'mdtraj/formats/xtc/'],
                    extra_compile_args=compiler_args)

    trr = Extension('mdtraj.formats.trr',
                    sources=['mdtraj/formats/xtc/src/xdrfile.c',
                             'mdtraj/formats/xtc/src/xdr_seek.c',
                             'mdtraj/formats/xtc/src/xdrfile_trr.c',
                             'mdtraj/formats/xtc/trr.pyx'],
                    include_dirs=['mdtraj/formats/xtc/include/',
                                  'mdtraj/formats/xtc/'],
                    extra_compile_args=compiler_args)

    zlib_include_dirs = []
    zlib_library_dirs = []
    if sys.platform == 'win32':
        # Conda puts the zlib headers in ./Library/... on windows
        # If you're not using conda, good luck!
        zlib_include_dirs += ["{}/Library/include".format(sys.prefix)]
        zlib_library_dirs += ["{}/Library/lib".format(sys.prefix)]
    else:
        # On linux (and mac(?)) these paths should work for a standard
        # install of python+zlib or a conda install of python+zlib
        zlib_include_dirs += ["{}/include".format(sys.prefix)]
        zlib_library_dirs += ["{}/lib".format(sys.prefix)]

    tng = Extension('mdtraj.formats.tng',
                    sources=glob('mdtraj/formats/tng/src/compression/*.c') +
                                ['mdtraj/formats/tng/src/lib/tng_io.c',
                                 'mdtraj/formats/tng/src/lib/md5.c',
                                 'mdtraj/formats/tng/tng.pyx'],
                    include_dirs=['mdtraj/formats/tng/include']
                                 + zlib_include_dirs,
                    define_macros=[('USE_ZLIB', 1)],
                    library_dirs=zlib_library_dirs,
                    libraries=['z'],
                    )

    dcd = Extension('mdtraj.formats.dcd',
                    sources=['mdtraj/formats/dcd/src/dcdplugin.c',
                             'mdtraj/formats/dcd/dcd.pyx'],
                    include_dirs=["mdtraj/formats/dcd/include/",
                                  'mdtraj/formats/dcd/'],
                    extra_compile_args=compiler_args)

    binpos = Extension('mdtraj.formats.binpos',
                       sources=['mdtraj/formats/binpos/src/binposplugin.c',
                                'mdtraj/formats/binpos/binpos.pyx'],
                       include_dirs=['mdtraj/formats/binpos/include/',
                                     'mdtraj/formats/binpos/'],
                       extra_compile_args=compiler_args)

    dtr = Extension('mdtraj.formats.dtr',
                    sources=['mdtraj/formats/dtr/src/dtrplugin.cxx',
                             'mdtraj/formats/dtr/dtr.pyx'],
                    include_dirs=['mdtraj/formats/dtr/include/',
                                  'mdtraj/formats/dtr/'],
                    define_macros=[('DESRES_READ_TIMESTEP2', 1)],
                    language='c++',
                    extra_compile_args=compiler_args,
                    libraries=extra_cpp_libraries)

    return [xtc, trr, tng, dcd, binpos, dtr]


def rmsd_extensions():
    compiler_args = (compiler.compiler_args_openmp + compiler.compiler_args_sse2 +
                     compiler.compiler_args_sse3 + compiler.compiler_args_opt +
                     compiler.compiler_args_warn)
    compiler_libraries = compiler.compiler_libraries_openmp

    libtheobald = StaticLibrary(
        'mdtraj.core.lib.libtheobald',
        sources=[
            'mdtraj/rmsd/src/theobald_rmsd.cpp',
            'mdtraj/rmsd/src/center.cpp'],
        include_dirs=[
            'mdtraj/rmsd/include'],
        export_include=['mdtraj/rmsd/include/theobald_rmsd.h',
                        'mdtraj/rmsd/include/center.h'],
        language="c++",
        # don't enable OpenMP
        extra_compile_args=(compiler.compiler_args_sse2 +
                            compiler.compiler_args_sse3 +
                            compiler.compiler_args_opt))

    rmsd = Extension('mdtraj._rmsd',
                     sources=[
                         'mdtraj/rmsd/src/theobald_rmsd.cpp',
                         'mdtraj/rmsd/src/rotation.cpp',
                         'mdtraj/rmsd/src/center.cpp',
                         'mdtraj/rmsd/_rmsd.pyx'],
                     include_dirs=['mdtraj/rmsd/include'],
                     extra_compile_args=compiler_args,
                     libraries=compiler_libraries,
                     language="c++")

    lprmsd = Extension('mdtraj._lprmsd',
                       sources=[
                           'mdtraj/rmsd/src/theobald_rmsd.cpp',
                           'mdtraj/rmsd/src/rotation.cpp',
                           'mdtraj/rmsd/src/center.cpp',
                           'mdtraj/rmsd/src/fancy_index.cpp',
                           'mdtraj/rmsd/src/Munkres.cpp',
                           'mdtraj/rmsd/src/euclidean_permutation.cpp',
                           'mdtraj/rmsd/_lprmsd.pyx'],
                       language='c++',
                       include_dirs=['mdtraj/rmsd/include'],
                       extra_compile_args=compiler_args,
                       libraries=compiler_libraries + extra_cpp_libraries)
    return rmsd, lprmsd, libtheobald


def geometry_extensions():
    compiler.initialize()
    compiler_args = (
        compiler.compiler_args_openmp +
        compiler.compiler_args_sse2 + compiler.compiler_args_sse3 +
        compiler.compiler_args_opt + compiler.compiler_args_warn)
    define_macros = None
    compiler_libraries = compiler.compiler_libraries_openmp + extra_cpp_libraries

    return [
        Extension('mdtraj.geometry._geometry',
            sources=['mdtraj/geometry/src/sasa.cpp',
                     'mdtraj/geometry/src/dssp.cpp',
                     'mdtraj/geometry/src/geometry.cpp',
                     'mdtraj/geometry/src/_geometry.pyx',],
            include_dirs=['mdtraj/geometry/include',
                          'mdtraj/geometry/src/kernels'],
            depends=['mdtraj/geometry/src/kernels/anglekernels.h',
                     'mdtraj/geometry/src/kernels/dihedralkernels.h',
                     'mdtraj/geometry/src/kernels/distancekernels.h'],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language='c++'),
        Extension('mdtraj.geometry.drid',
            sources=["mdtraj/geometry/drid.pyx",
                     "mdtraj/geometry/src/dridkernels.cpp",
                     "mdtraj/geometry/src/moments.cpp"],
            include_dirs=["mdtraj/geometry/include"],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language='c++'),
        Extension('mdtraj.geometry.neighbors',
            sources=["mdtraj/geometry/neighbors.pyx",
                     "mdtraj/geometry/src/neighbors.cpp"],
            include_dirs=["mdtraj/geometry/include",],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language='c++'),
        Extension('mdtraj.geometry.neighborlist',
            sources=["mdtraj/geometry/neighborlist.pyx",
                     "mdtraj/geometry/src/neighborlist.cpp"],
            include_dirs=["mdtraj/geometry/include",],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language='c++'),
        ]


write_version_py(VERSION, ISRELEASED, 'mdtraj/version.py')

metadata = \
    dict(name='mdtraj',
      author='Robert McGibbon',
      author_email='rmcgibbo@gmail.com',
      description=DOCLINES[0],
      long_description="\n".join(DOCLINES[2:]),
      version=__version__,
      license='LGPLv2.1+',
      url='http://mdtraj.org',
      download_url = "https://github.com/rmcgibbo/mdtraj/releases/latest",
      platforms=['Linux', 'Mac OS-X', 'Unix', 'Windows'],
      classifiers=CLASSIFIERS.splitlines(),
      packages=find_packages(),
      cmdclass={'build_ext': build_ext},
      install_requires=['numpy>=1.6',
                        'scipy',
                        'astunparse',
                        'pyparsing',
                        ],
      package_data={'mdtraj.formats.pdb': ['data/*'], },
      zip_safe=False,
      entry_points={'console_scripts':
          ['mdconvert = mdtraj.scripts.mdconvert:entry_point',
           'mdinspect = mdtraj.scripts.mdinspect:entry_point']},
)


if __name__ == '__main__':
    # Don't use numpy if we are just - non-build actions are required to succeed
    # without NumPy for example when pip is used to install Scipy when
    # NumPy is not yet present in the system.
    run_build = parse_setuppy_commands()
    if run_build:
        extensions = format_extensions()
        extensions.extend(rmsd_extensions())
        extensions.extend(geometry_extensions())

        # most extensions use numpy, add headers for it.
        try:
            import Cython as _c
            from Cython.Build import cythonize
            if _c.__version__ < '0.29':
                raise ImportError("Too old")
        except ImportError as e:
            print('mdtrajs setup depends on Cython (>=0.29). Install it prior invoking setup.py')
            print(e)
            sys.exit(1)
        try:
            import numpy as np
        except ImportError:
            print('mdtrajs setup depends on NumPy. Install it prior invoking setup.py')
            sys.exit(1)

        for e in extensions:
            e.include_dirs.append(np.get_include())
        metadata['ext_modules'] = cythonize(extensions, language_level=sys.version_info[0])

    setup(**metadata)
